# flake8: noqa
from typing import Optional

import requests

from arthur_bench.client.http.requests import HTTPClient
from arthur_bench.client.rest.admin.client import ArthurAdminClient
from arthur_bench.client.rest.admin.models import LoginRequest, AuthenticationInfo
from arthur_bench.exceptions import (
    UserValueError,
    ForbiddenError,
    UnauthorizedError,
    NotFoundError,
    ArthurInternalError,
)


def get_current_org(
    api_http_host: str, auth_token: str, verify_ssl: bool = True
) -> Optional[str]:
    """Get the current organization for the provided access key

    :param api_http_host: base url of the host to connect to, including protocol
    (e.g. "https://app.arthur.ai")
    :param auth_token: API Key to pass to the API
    :param verify_ssl: Boolean for whether requests should verify that the SSL
      certificate is valid and not self-signed
    :return: the organization ID associated with the provided access key, None if no
      such organization exists
    """
    try:
        internal_user_org = get_arthur_internal_user_org(
            api_http_host, auth_token, verify_ssl
        )
    except ForbiddenError:
        auth_info = get_auth_info(api_http_host, auth_token, verify_ssl)
        if auth_info is None:
            raise ArthurInternalError("Invalid / non-existent auth_info")
        elif len(auth_info.organization_ids) == 1:
            # If the user is authenticated into a single org, automatically select it as
            #  the user's current org
            return auth_info.organization_ids[0]
        elif len(auth_info.organization_ids) > 1:
            # Raise exception and give the user options of available current
            #  organizations
            authenticated_org_id_str = "".join(
                f"{org_id}\n" for org_id in auth_info.organization_ids
            )[:-1]
            raise UserValueError(
                f"Your access_key provides access to multiple organizations - "
                "please specify one of the following: " + authenticated_org_id_str
            )
        else:  # len(auth_info.organization_ids) == 0
            return None

    # Valid internal_user_org was found
    return internal_user_org


def get_auth_info(
    api_http_host: str, auth_token: str, verify_ssl: bool = True
) -> AuthenticationInfo:
    """Get the AuthInfo struct associated with the provided access key

    :param api_http_host: base url of the host to connect to, including protocol
    (e.g. "https://app.arthur.ai")
    :param auth_token: Token to fetch authentication info for
    :param verify_ssl: Boolean for whether requests should verify that the SSL
    certificate is valid and not self-signed
    :return: the AuthInfo associated with the provided access key
    :permissions: N/A
    """
    admin_client = ArthurAdminClient(
        HTTPClient(
            base_url=api_http_host,
            verify_ssl=verify_ssl,
            default_headers={"Authorization": auth_token},
        )
    )
    try:
        auth_info = admin_client.authenticate()
    except requests.exceptions.SSLError as e:
        raise UserValueError(
            f"""SSL Error connecting to {api_http_host}, please connect to a secure 
            server or use verify_ssl=False to override security checks"""
        ) from e
    except requests.RequestException as e:
        raise UserValueError(
            f"Failed to connect to {api_http_host}, please ensure the URL is correct"
        ) from e
    except NotFoundError as e:
        raise ArthurInternalError(
            "Auth Info endpoint not implemented by api-host"
        ) from e

    return auth_info


def get_arthur_internal_user_org(
    api_http_host: str, auth_token: str, verify_ssl: bool = True
) -> Optional[str]:
    """Get the current organization for the provided Arthur auth token belonging to an
      Arthur internal user

    :param api_http_host: base url of the host to connect to, including protocol
      (e.g. "https://app.arthur.ai")
    :param auth_token: auth token to pass to the API
    :param verify_ssl: if True, verify that the SSL certificate is valid and not
      self-signed
    :return: the organization ID associated with the provided access key, None if no
      such organization exists
    :permissions: N/A
    """
    admin_client = ArthurAdminClient(
        HTTPClient(
            base_url=api_http_host,
            verify_ssl=verify_ssl,
            default_headers={"Authorization": auth_token},
        )
    )
    try:
        user = admin_client.get_current_user()
    except requests.exceptions.SSLError as e:
        raise UserValueError(
            f"""SSL Error connecting to {api_http_host}, please connect to a secure
              server or use verify_ssl=False to override security checks"""
        ) from e
    except requests.RequestException as e:
        raise UserValueError(
            f"Failed to connect to {api_http_host}, please ensure the URL is correct"
        ) from e
    except UnauthorizedError as e:
        raise UserValueError(
            "Unauthorized, please ensure your access information is correct"
        ) from e

    return user.organization_id


def user_login(
    api_http_host: str, login: str, password: str, verify_ssl: bool = True
) -> str:
    """Static convenience function to get a new auth token for the provided username and
      password

    :param api_http_host: base url of the host to connect to, including protocol
      (e.g. "https://app.arthur.ai")
    :param login: the username or password to use to log in
    :param password: password for the user
    :param verify_ssl: Boolean for whether requests should verify that the SSL
      certificate is valid and not self-signed
    :return: an access_key
    """

    admin_client = ArthurAdminClient(
        HTTPClient(base_url=api_http_host, verify_ssl=verify_ssl)
    )
    try:
        _, login_cookies = admin_client.login(
            LoginRequest(login=login, password=password)
        )
    # TODO: move all these checkings into the HTTPClient
    except requests.exceptions.SSLError as e:
        raise UserValueError(
            f"""SSL Error connecting to {api_http_host}, please connect to a secure
              server or use verify_ssl=False to override security checks"""
        ) from e
    except requests.RequestException as e:
        raise UserValueError(
            f"Failed to connect to {api_http_host}, please ensure the URL is correct"
        ) from e
    except ForbiddenError as e:
        raise UserValueError(
            f"Unauthorized, please ensure your username and password are correct"
        ) from e

    auth_token = login_cookies.get("Authorization")

    return auth_token



from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import List, Optional
from uuid import UUID

from pydantic import BaseModel, Field, validator


# COMMON


class ScoringMethodType(str, Enum):
    BuiltIn = "built_in"
    Custom = "custom"


class ScoringMethod(BaseModel):
    name: str
    type: ScoringMethodType
    config: dict = {}


# REQUESTS


class TestCaseRequest(BaseModel):
    """
    An input, reference output pair.
    """

    input: str
    """
    Input to the test case. Does not include the prompt template.
    """
    reference_output: Optional[str]
    """
    Reference or "Golden" output for the given input.
    """


class TestSuiteRequest(BaseModel):
    """
    Test case data and metadata for the test suite.
    """

    name: str
    description: Optional[str] = None
    scoring_method: ScoringMethod
    test_cases: List[TestCaseRequest] = Field(..., min_items=1)

    @validator("test_cases")
    def null_reference_outputs_all_or_none(cls, v):
        last_ref_output_null = None
        for tc in v:
            # get ref output value
            if isinstance(tc, TestCaseRequest):
                ref_val = tc.reference_output
            elif isinstance(tc, dict):
                ref_val = tc.get("reference_output", None)
            else:
                raise TypeError(
                    f"Unable to extract reference output value for type '{type(v)}'"
                )

            # check it matches what we've been seeing
            if ref_val is None and last_ref_output_null is False:
                raise ValueError(
                    "Test Suite has both null and non-null reference outputs. Reference"
                    " outputs for test cases within a suite should be all null or all"
                    " non-null."
                )
            if ref_val is not None and last_ref_output_null is True:
                raise ValueError(
                    "Test Suite has both null and non-null reference outputs. Reference"
                    " outputs for test cases within a suite should be all null or all"
                    " non-null."
                )
            if ref_val is None:
                last_ref_output_null = True
            else:
                last_ref_output_null = False

        return v

    @validator("scoring_method", pre=True)
    def scoring_method_backwards_compatible(cls, v):
        if isinstance(v, str):
            return ScoringMethod(name=v, type=ScoringMethodType.BuiltIn)
        return v


class TestCaseOutput(BaseModel):
    """
    A generated output, score pair
    """

    id: UUID
    """
    Optional unique identifier for this test case of the suite and run
    """
    output: str
    """
    Generated output for test case
    """
    score: float
    """
    Score assigned to output
    """


class CreateRunRequest(BaseModel):
    name: str
    """
    Name identifier of the run
    """
    test_cases: List[TestCaseOutput] = Field(alias="test_case_outputs")
    """
    List of outputs and scores for all cases in the test suite
    """
    description: Optional[str] = None
    """
    Optional description of the run
    """
    model_name: Optional[str] = None
    """
    Optional model name identifying the model used to generate outputs
    """
    foundation_model: Optional[str] = None
    """
    Optional foundation model name identifiying the pretrained model used to generate
    outputs
    """
    prompt_template: Optional[str] = None
    """
    Optional prompt template name identifying the global prompt used to generate outputs
    """
    model_version: Optional[str] = None
    """
    Optional model version identifying the version of the model used to generate outputs
    """

    class Config:
        allow_population_by_field_name = True


# RESPONSES


class TestSuiteMetadata(BaseModel):
    id: UUID
    name: str
    scoring_method: ScoringMethod
    last_run_time: Optional[datetime] = None
    description: Optional[str] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None


class PaginatedTestSuites(BaseModel):
    test_suites: List[TestSuiteMetadata]
    page: int
    page_size: int
    total_pages: int
    total_count: int


class TestCaseResponse(BaseModel):
    id: UUID
    input: str
    """
    Input to the test case. Does not include the prompt template.
    """
    reference_output: Optional[str] = None
    """
    Reference or "Golden" output for the given input.
    """


class PaginatedTestSuite(BaseModel):
    id: UUID
    name: str
    scoring_method: ScoringMethod
    test_cases: List[TestCaseResponse]
    created_at: datetime
    updated_at: datetime
    description: Optional[str] = None
    last_run_time: Optional[datetime] = None
    num_runs: int = 0
    page: Optional[int] = None
    page_size: Optional[int] = None
    total_pages: Optional[int] = None
    total_count: Optional[int] = None


class TestRunMetadata(BaseModel):
    id: UUID
    name: str
    created_at: datetime
    updated_at: datetime
    avg_score: Optional[float] = None
    model_version: Optional[str] = None
    prompt_template: Optional[str] = None


class PaginatedRuns(BaseModel):
    """
    Paginated list of runs for a test suite.
    """

    test_runs: List[TestRunMetadata]
    page: int
    page_size: int
    total_pages: int
    total_count: int


class HistogramItem(BaseModel):
    count: int
    low: float
    high: float


class SummaryItem(BaseModel):
    id: UUID
    name: str
    avg_score: float
    histogram: List[HistogramItem]


class TestSuiteSummary(BaseModel):
    summary: List[SummaryItem]
    page: int
    page_size: int
    total_pages: int
    total_count: int
    num_test_cases: int


class CreateRunResponse(BaseModel):
    id: UUID


class RunResult(BaseModel):
    id: UUID
    output: str
    score: float
    input: Optional[str] = None
    reference_output: Optional[str] = None


class PaginatedRun(BaseModel):
    """
    Paginated list of prompts, reference outputs, and model outputs for a particular
    run.
    """

    id: UUID
    name: str
    test_suite_id: UUID
    test_cases: List[RunResult] = Field(alias="test_case_runs")
    updated_at: datetime
    created_at: datetime
    page: Optional[int] = None
    page_size: Optional[int] = None
    total_pages: Optional[int] = None
    total_count: Optional[int] = None

    class Config:
        allow_population_by_field_name = True



import os

from arthur_bench.exceptions import UserValueError, MissingParameterError
from arthur_bench.client.bench_client import BenchClient
from arthur_bench.client.local import LocalBenchClient
from arthur_bench.client.rest import ArthurClient


def _get_bench_client() -> BenchClient:
    client: BenchClient
    url = os.getenv("ARTHUR_API_URL")
    if url:  # if remote url is specified use remote client
        try:
            client = ArthurClient(url=url).bench
        except (UserValueError, MissingParameterError) as e:
            raise UserValueError(
                f"You must provide authentication when using remote url: {e}"
            )
    else:
        client = LocalBenchClient()
    return client



from typing import Optional, Dict, cast, List, Any
from http import HTTPStatus

# import http client
from arthur_bench.client.http.requests import HTTPClient
from arthur_bench.client.bench_client import BenchClient

from arthur_bench.models.models import (
    PaginatedTestSuites,
    PaginatedRun,
    CreateRunRequest,
    PaginatedRuns,
    TestSuiteRequest,
    PaginatedTestSuite,
    TestSuiteSummary,
    CreateRunResponse,
)

from arthur_bench.models.scoring import (
    HallucinationScoreRequest,
    HallucinationScoreResponse,
)


PATH_PREFIX = "/api/v3"


class ArthurBenchClient(BenchClient):
    """
    A Python client to interact with the Arthur Bench API
    """

    def __init__(self, http_client: HTTPClient):
        """
        Create a new ArthurBenchClient from an HTTPClient

        :param http_client: the :class:`~arthurai.client.http.requests.HTTPClient`
            to use for underlying requests
        """
        self.http_client = http_client
        self.http_client.set_path_prefix(PATH_PREFIX)

    def get_test_suites(
        self,
        name: Optional[str] = None,
        sort: Optional[str] = None,
        scoring_method: Optional[List[str]] = None,
        page: int = 1,
        page_size: int = 5,
    ) -> PaginatedTestSuites:
        """
                Gets test suites

                Sort by latest run by default.
        If `name` query parameter is provided, filter on test suite name.

                :param name:
                :param sort:
                :param scoring_method:
        """

        params: Dict[str, Any] = {}
        if name is not None:
            params["name"] = name
        if sort is not None:
            params["sort"] = sort
        if scoring_method is not None:
            params["scoring_method"] = scoring_method
        if page is not None:
            params["page"] = page  # type: ignore
        if page_size is not None:
            params["page_size"] = page_size  # type: ignore

        parsed_resp = cast(
            Dict,
            self.http_client.get(
                "/bench/test_suites",
                params=params,
                validation_response_code=HTTPStatus.OK,
            ),
        )
        return PaginatedTestSuites(**parsed_resp)

    def create_test_suite(self, json_body: TestSuiteRequest) -> PaginatedTestSuite:
        """
        Creates a new test suite from reference data using specified scoring_method for
        scoring

        :param json_body:
        """

        parsed_resp = cast(
            Dict,
            self.http_client.post(
                "/bench/test_suites",
                json=json_body.json(
                    exclude={
                        "created_at": True,
                        "created_by": True,
                        "bench_version": True,
                        # TODO: add REST data store support for scoring method config
                        "scoring_method": {"config"},
                    }
                ),
                validation_response_code=HTTPStatus.CREATED,
            ),
        )
        return PaginatedTestSuite(**parsed_resp)

    def get_test_suite(
        self,
        test_suite_id: str,
        page: int = 1,
        page_size: int = 5,
    ) -> PaginatedTestSuite:
        """
        Get reference data for an existing test suite

        :param test_suite_id:
        """
        params = {}
        if page is not None:
            params["page"] = page  # type: ignore
        if page_size is not None:
            params["page_size"] = page_size  # type: ignore

        parsed_resp = cast(
            Dict,
            self.http_client.get(
                f"/bench/test_suites/{test_suite_id}",
                params=params,
                validation_response_code=HTTPStatus.OK,
            ),
        )
        return PaginatedTestSuite(**parsed_resp)

    def get_summary_statistics(
        self,
        test_suite_id: str,
        run_id: Optional[str] = None,
        page: int = 1,
        page_size: int = 5,
    ) -> TestSuiteSummary:
        """
        Get paginated summary statistics of a test suite

        Defaults to page size of 5.

        :param test_suite_id:
        :param run_id:
        :param page:
        :param page_size:
        """

        params = {}
        if run_id is not None:
            params["run_id"] = run_id
        if page is not None:
            params["page"] = page  # type: ignore
        if page_size is not None:
            params["page_size"] = page_size  # type: ignore

        parsed_resp = cast(
            Dict,
            self.http_client.get(
                f"/bench/test_suites/{test_suite_id}/runs/summary",
                params=params,
                validation_response_code=HTTPStatus.OK,
            ),
        )
        return TestSuiteSummary(**parsed_resp)

    def get_runs_for_test_suite(
        self,
        test_suite_id: str,
        sort: Optional[str] = None,
        page: int = 1,
        page_size: int = 5,
    ) -> PaginatedRuns:
        """
        Get runs for a particular test suite (identified by test_suite_id)

        :param test_suite_id:
        :param sort:
        """

        params = {}
        if sort is not None:
            params["sort"] = sort
        if page is not None:
            params["page"] = page  # type: ignore
        if page_size is not None:
            params["page_size"] = page_size  # type: ignore

        parsed_resp = cast(
            Dict,
            self.http_client.get(
                f"/bench/test_suites/{test_suite_id}/runs",
                params=params,
                validation_response_code=HTTPStatus.OK,
            ),
        )
        return PaginatedRuns(**parsed_resp)

    def create_new_test_run(
        self, test_suite_id: str, json_body: CreateRunRequest
    ) -> CreateRunResponse:
        """
        Creates a new test run with model version / associated metadata


        :param test_suite_id:
        :param json_body:
        """

        parsed_resp = cast(
            Dict,
            self.http_client.post(
                f"/bench/test_suites/{test_suite_id}/runs",
                json=json_body.json(by_alias=True),
                validation_response_code=HTTPStatus.CREATED,
            ),
        )
        return CreateRunResponse(**parsed_resp)

    def get_test_run(
        self,
        test_suite_id: str,
        test_run_id: str,
        page: int = 1,
        page_size: int = 5,
        sort: Optional[bool] = None,
    ) -> PaginatedRun:
        """
        Get a test run with input, output, and reference data

        :param test_suite_id:
        :param test_run_id:
        :param page:
        :param page_size:
        :param sort:
        """

        params = {}
        if page is not None:
            params["page"] = page  # type: ignore
        if page_size is not None:
            params["page_size"] = page_size  # type: ignore
        if sort is not None:
            params["sort"] = sort

        parsed_resp = cast(
            Dict,
            self.http_client.get(
                f"/bench/test_suites/{test_suite_id}/runs/{test_run_id}",
                params=params,
                validation_response_code=HTTPStatus.OK,
            ),
        )
        return PaginatedRun(**parsed_resp)

    def delete_test_suite(self, test_suite_id: str):
        """
        Deletes test suite

        Is idempotent.

        :param test_suite_id:
        """

        raw_resp = self.http_client.delete(
            f"/bench/test_suites/{test_suite_id}",
            validation_response_code=HTTPStatus.NO_CONTENT,
            return_raw_response=True,
        )
        return raw_resp

    def delete_test_run(self, test_suite_id: str, test_run_id: str):
        """
        Deletes a test run

        Is idempotent.

        :param test_suite_id:
        :param test_run_id:
        """

        raw_resp = self.http_client.delete(
            f"/bench/test_suites/{test_suite_id}/runs/{test_run_id}",
            validation_response_code=HTTPStatus.NO_CONTENT,
            return_raw_response=True,
        )
        return raw_resp

    def score_hallucination(
        self, json_body: HallucinationScoreRequest
    ) -> HallucinationScoreResponse:
        parsed_resp = cast(
            Dict,
            self.http_client.post(
                "/bench/scoring/hallucination",
                json=json_body.json(),
                validation_response_code=HTTPStatus.CREATED,
            ),
        )
        return HallucinationScoreResponse(**parsed_resp)



from typing import List, Optional

from langchain.chains import LLMChain
from langchain.chat_models import ChatOpenAI

from arthur_bench.exceptions import UserValueError
from arthur_bench.scoring import Scorer
from arthur_bench.scoring.prompts.qa_correctness import DECIDE


class QAQualityCorrectness(Scorer):
    """
    Given an input question, context string, and model generation, determine if the
    generation produced a correct answer.
    """

    def __init__(self):
        self.evaluator = LLMChain(llm=ChatOpenAI(temperature=0), prompt=DECIDE)

    @staticmethod
    def name() -> str:
        return "qa_correctness"

    @staticmethod
    def requires_reference() -> bool:
        return False

    def to_dict(self, warn=False):
        return {}

    def run_batch(
        self,
        candidate_batch: List[str],
        reference_batch: Optional[List[str]] = None,
        input_text_batch: Optional[List[str]] = None,
        context_batch: Optional[List[str]] = None,
    ) -> List[float]:
        """
        Reference batch is not used for this scoring method, QA correctness requires an
        input_text_batch and context_batch
        """
        if input_text_batch is None:
            raise UserValueError(
                "input text is required for this scoring method. Please provide a "
                "dataframe column or a list of your input text strings in the Test "
                "Suite."
            )
        if context_batch is None:
            raise UserValueError(
                "context is required for this scoring method. Please provide a "
                "dataframe column or a list of your context strings in the Test Suite."
            )

        res = []
        for i in range(len(input_text_batch)):
            llmchoice = self.evaluator(
                {
                    "question": input_text_batch[i],
                    "context": context_batch[i],
                    "answer": candidate_batch[i],
                }
            )["text"]
            if llmchoice not in ["0", "1", "NA"]:
                llmchoice = "-1"
            llmchoice = {"0": 0.0, "1": 1.0, "NA": -1.0, "-1": -1.0}[llmchoice]
            res.append(llmchoice)
        return res



# flake8: noqa

from langchain.prompts.chat import (
    ChatPromptTemplate,
    SystemMessagePromptTemplate,
    AIMessagePromptTemplate,
    HumanMessagePromptTemplate,
)

system_message_prompt = SystemMessagePromptTemplate.from_template(
    """You compare two summaries of a text. You respond with a Choice, either a 0, 1, or tie ONLY.
(0 = response 0 is better, 1 = response 1 is better, tie = no significant difference between the responses).
A good summary captures the most important information in the text and doesnt focus too much on small details.
A bad summary has information that is conflicting or irrelevant to the original text, or has typos of words in the text."""
)
example_summaries_1 = HumanMessagePromptTemplate.from_template(
    """You compare two summaries of a text. You respond with a Choice, either a 0, 1, or tie ONLY.
(0 = response 0 is better, 1 = response 1 is better, tie = no significant difference between the responses).
A good summary captures the most important information in the text and doesnt focus too much on small details.
A bad summary has information that is conflicting or irrelevant to the original text, or has typos of words in the text.
Text: (The Hollywood Reporter)Add another fan-favorite
character to the cast of next year's X-Men: Apocalypse, with director Bryan Singer announcing
 via Instagram that Olivia Munn will play the telepathic Psylocke in the follow-up to X-Men:
Days of Future Past. Singer revealed that the Newsroom actress would play Betsy Braddock in
the movie (presumably before the confusing and complicated plot twist that saw Psylocke
change from a Caucasian former supermodel to a Japanese ninja for no immediately obvious
reason).
Response 0:
Bryan Singer said that he would love to see Olivio Mun in x-men: apocalypse. 
Response 1:
Bryan Singer has revealed that Olivia Munn will star as Psylocke in next year's x-men:
apocalypse - a character created more than 20 years ago for the x-men.
Choice:"""
)
example_choice_1 = AIMessagePromptTemplate.from_template("1")
example_summaries_2 = HumanMessagePromptTemplate.from_template(
    """You compare two summaries of a text. You respond with a Choice, either a 0, 1, or tie ONLY.
(0 = response 0 is better, 1 = response 1 is better, tie = no significant difference between the responses).
A good summary captures the most important information in the text and doesnt focus too much on small details.
A bad summary has information that is conflicting or irrelevant to the original text, or has typos of words in the text.
Text: (The Hollywood Reporter)Add another fan-favorite
character to the cast of next year's X-Men: Apocalypse, with director Bryan Singer announcing
 via Instagram that Olivia Munn will play the telepathic Psylocke in the follow-up to X-Men:
Days of Future Past. Singer revealed that the Newsroom actress would play Betsy Braddock in
the movie (presumably before the confusing and complicated plot twist that saw Psylocke
change from a Caucasian former supermodel to a Japanese ninja for no immediately obvious
reason).
Response 0:
Bryan Singer announced Olivia Munn will lead as the classic character Psylocke in the upcoming movie X-Men: Apocalypse
Response 1:
Bryan Singer has revealed that Olivia Munn will star as Psylocke in next year's x-men:
apocalypse - a character created more than 20 years ago for the x-men.
Choice:"""
)
example_choice_2 = AIMessagePromptTemplate.from_template("tie")
comparison_template = HumanMessagePromptTemplate.from_template(
    """You compare two summaries of a text. You respond with a Choice, either a 0, 1, or tie ONLY.
(0 = response 0 is better, 1 = response 1 is better, tie = no significant difference between the responses).
A good summary captures the most important information in the text and doesnt focus too much on small details.
A bad summary has information that is conflicting or irrelevant to the original text, or has typos of words in the text.
Text: {text}
Response 0: {summary_A}
Response 1: {summary_B}
Choice:"""
)

COMPARE = ChatPromptTemplate.from_messages(
    [
        system_message_prompt,
        example_summaries_1,
        example_choice_1,
        example_summaries_2,
        example_choice_2,
        comparison_template,
    ]
)



import logging
from urllib.parse import urlparse

from arthur_bench.exceptions import UserValueError


logger = logging.getLogger(__name__)


def construct_url(*parts: str, validate=True, default_https=True) -> str:
    """Construct a URL from various parts

    Useful for joining pieces which may or may not have leading and/or trailing
    slashes. e.g. construct_url("https://arthur.ai/", "/api/v3", "/users") will yield
    the same valid url as construct_url("https://arthur.ai", "api/v3/", "users/"):
    "https://arthur.ai/api/v3/users".

    :param validate: if True, validate that the URL is valid
    :param default_https: if True, allow urls without a scheme and use https by default
    :param parts: strings from which to construct the url
    :return: a fully joined url, with NO trailing slash
    """
    # join parts
    url = "/".join(s.strip("/") for s in parts)

    # add scheme
    parsed_url = urlparse(url)
    if parsed_url.scheme is None or parsed_url.scheme == "":
        if default_https:
            logger.warning("No url scheme provided, defaulting to https")
            url = "https://" + url
            parsed_url = urlparse(url)
        elif validate:
            raise UserValueError(f"No scheme provided in URL {url}")

    # validate
    if validate and (
        parsed_url.scheme is None
        or parsed_url.scheme == ""
        or parsed_url.netloc is None
        or parsed_url.netloc == ""
    ):
        joiner = "', '"
        raise UserValueError(
            f"Invalid url, cannot construct URL from parts '{joiner.join(parts)}'"
        )

    return url



import json
import uuid
from pathlib import Path
from pydantic import BaseModel


class TelemetryConfig(BaseModel):
    user_id: str
    log_notice_of_usage_data: bool
    push_usage_data: bool


def _get_config_file_name() -> Path:
    return Path("~/.bench_config/config.json").expanduser()


def get_or_persist_id() -> TelemetryConfig:
    config_file = _get_config_file_name()
    # Create directory if it does not exist.
    config_file.parent.mkdir(exist_ok=True)

    # File exists.  Return its contents.
    if config_file.exists():
        return TelemetryConfig(**json.loads(config_file.open().read()))

    # File doesn't exist.  Generate and return config.
    id = uuid.uuid4()
    new_config = TelemetryConfig(
        user_id=str(id), log_notice_of_usage_data=False, push_usage_data=True
    )
    config_file.open("w+").write(new_config.json())

    # Indicate that file was just created to the caller => we should log notice.
    new_config.log_notice_of_usage_data = True
    return new_config


def persist_usage_data(push_usage_data: bool):
    config_file = _get_config_file_name()
    # Create directory if it does not exist.
    config_file.parent.mkdir(exist_ok=True)

    msg = f"Successfully set Bench anonymous usage data collection to {push_usage_data}"

    if config_file.exists():
        cfg = TelemetryConfig(**json.loads(config_file.open().read()))
        cfg.push_usage_data = push_usage_data
        config_file.open("w+").write(cfg.json())
        print(msg)
        return

    id = uuid.uuid4()
    new_config = TelemetryConfig(
        user_id=str(id), log_notice_of_usage_data=False, push_usage_data=push_usage_data
    )
    config_file.open("w+").write(new_config.json())
    print(msg)



from abc import abstractmethod, ABC
import sys
import json
import logging
from typing import List, Optional, TypeVar, get_origin, get_args, Union
from inspect import signature, Parameter

from tqdm import tqdm
from arthur_bench.models.models import ScoringMethodType


logger = logging.getLogger(__name__)


SINGLE_ITEM_BATCH_DEFAULT = 1


TScorer = TypeVar("TScorer", bound="Scorer")


def _can_omit(parameter: Parameter):
    is_optional = get_origin(parameter.annotation) is Union and type(None) in get_args(
        parameter.annotation
    )
    is_kwargs = parameter.name == "args" or parameter.name == "kwargs"
    return is_optional or is_kwargs


class Scorer(ABC):
    """
    Base class for all scorers.
    """

    @staticmethod
    @abstractmethod
    def name() -> str:
        """
        Get the name of this Scorer
        :return: the Scorer name
        """
        raise NotImplementedError

    @staticmethod
    def requires_reference() -> bool:
        return True

    @abstractmethod
    def run_batch(
        self,
        candidate_batch: List[str],
        reference_batch: Optional[List[str]] = None,
        input_text_batch: Optional[List[str]] = None,
        context_batch: Optional[List[str]] = None,
    ) -> List[float]:
        """
        Score a batch of candidate generations.

        :param candidate_batch: candidate generations to score
        :param reference_batch: reference strings representing target outputs
        :param input_text_batch: optional corresponding inputs
        :param context_batch: optional corresponding contexts, if needed by scorer
        """
        raise NotImplementedError

    def run(
        self,
        candidate_outputs: List[str],
        reference_outputs: Optional[List[str]] = None,
        inputs: Optional[List[str]] = None,
        contexts: Optional[List[str]] = None,
        batch_size: int = SINGLE_ITEM_BATCH_DEFAULT,
    ) -> List[float]:
        """
        Score a set of test cases. This method doesn't need to be implemented in most
        cases, but can be overriden to add additional functionality such as
        task-specific logging.

        :param candidate_outputs: candidate generations to score
        :param reference_outputs: reference strings representing target outputs
        :param inputs: input strings being tested
        :param contexts: optional corresponding contexts, if needed by scorer
        :param batch_size: size of batches
        :return: array of scores for batch
        """
        all_scores = []
        with tqdm(total=len(candidate_outputs)) as pbar:
            for i in range(0, len(candidate_outputs), batch_size):
                input_batch = (
                    list(inputs[i : i + batch_size]) if inputs is not None else None
                )
                ref_batch = (
                    list(reference_outputs[i : i + batch_size])
                    if reference_outputs is not None
                    else None
                )

                context_batch = (
                    None if contexts is None else contexts[i : i + batch_size]
                )
                scores = self.run_batch(
                    candidate_outputs[i : i + batch_size],
                    ref_batch,
                    input_batch,
                    context_batch,
                )

                all_scores.extend(scores)
                pbar.update(len(candidate_outputs))

        return all_scores

    def to_dict(self, warn=False):
        """
        Provides a json serializable representation of the scorer.
        """
        config = self.__dict__
        valid_args = [p for p in signature(self.__init__).parameters.values()]

        # warn if arguments missing from initialization for reloading
        for arg in valid_args:
            if not _can_omit(arg) and arg.name not in config:
                if warn:
                    logger.warning(
                        f"scorer requires argument {arg} but argument is not included "
                        "in json representation. this may effect test suite reloading, "
                        "consider implementing custom to_dict and from_dict methods"
                    )

        jsonable_config = {}
        # remove non serializable args
        for key, val in config.items():
            try:
                _ = json.dumps(val)
                jsonable_config[key] = val
            except TypeError:
                if warn:
                    logger.warning(
                        f"not including attribute {key} in config as it is not json"
                        "serializable. consider implementing custom to_dict and "
                        "from_dict methods"
                    )
        return jsonable_config

    @classmethod
    def from_dict(cls, config: dict):
        """
        Load a scorer from a json configuration file.
        """
        valid_args = [p for p in signature(cls.__init__).parameters.keys()]

        init_config = {}
        for key, value in config.items():
            if key in valid_args:
                init_config[key] = value
        return cls(**init_config)

    @classmethod
    def type(cls) -> ScoringMethodType:
        """
        Supplies whether a scorer is built-in or custom.

        This method is implemented by checking whether the Scorer class is part of the
        `arthur_bench.scoring` module.
        :return: the type (built-in or custom)
        """
        try:
            module = sys.modules[cls.__module__].__file__
            if module is not None and "arthur_bench/scoring" in module:
                return ScoringMethodType.BuiltIn
            else:
                return ScoringMethodType.Custom
        except AttributeError:
            return ScoringMethodType.Custom



from collections import defaultdict
import pytest
import pandas as pd
from unittest import mock
from typing import List, Optional

from tests.helpers import (
    assert_test_suite_equal,
    assert_test_run_equal,
    get_mock_client,
)
from arthur_bench.run.testsuite import TestSuite
from arthur_bench.scoring import Scorer
from tests.fixtures.mock_requests import MOCK_SUITE, MOCK_SUITE_CUSTOM, MOCK_RUN
from tests.fixtures.mock_responses import (
    MOCK_SUITE_RESPONSE_WITH_PAGES,
    MOCK_SUITE_CUSTOM_RESPONSE_WITH_PAGES,
)
from tests.fixtures.mock_data import (
    MOCK_DATAFRAME,
    MOCK_INPUTS,
    MOCK_REFERENCE_OUTPUTS,
    MOCK_CUSTOM_DATAFRAME,
    MOCK_OUTPUTS,
)


@pytest.fixture
def mock_client():
    return get_mock_client()


@pytest.fixture
def test_suite_default(mock_load_scoring, mock_client):
    with mock.patch("arthur_bench.run.testsuite._initialize_scorer", mock_load_scoring):
        return TestSuite(
            name="test_suite",
            scoring_method="bertscore",
            reference_data=MOCK_DATAFRAME,
            client=mock_client,
        )


class MockScoringMethod(Scorer):
    @staticmethod
    def name():
        return "bertscore"

    @staticmethod
    def type():
        return "built_in"

    def run_batch(
        self,
        reference_batch: List[str],
        candidate_batch: List[str],
        input_text_batch: Optional[List[str]] = None,
        context_batch: Optional[List[str]] = None,
    ) -> List[float]:
        return [0.9, 0.7]


class CustomScorer(Scorer):
    def __init__(self, custom_name="param_name"):
        self.custom_name = custom_name

    @staticmethod
    def name():
        return "test_custom_scorer"

    def run_batch(
        self,
        reference_batch: List[str],
        candidate_batch: List[str],
        input_text_batch: Optional[List[str]] = None,
        context_batch: Optional[List[str]] = None,
    ) -> List[float]:
        return [0.5 for _ in range(len(reference_batch))]


@pytest.fixture(scope="session")
def mock_load_scoring():
    def load_scoring_method(scoring_method_arg):
        return MockScoringMethod()

    return load_scoring_method


@pytest.mark.parametrize(
    "params,expected",
    [
        (
            {
                "name": "test_suite",
                "scoring_method": "bertscore",
                "reference_data": MOCK_DATAFRAME,
            },
            MOCK_SUITE,
        ),
        (
            {
                "name": "test_suite_custom",
                "scoring_method": CustomScorer(),
                "description": "test_description",
                "reference_data": MOCK_CUSTOM_DATAFRAME,
                "input_column": "custom_prompt",
                "reference_column": "custom_reference",
            },
            MOCK_SUITE_CUSTOM,
        ),
        (
            {
                "name": "test_suite",
                "scoring_method": "bertscore",
                "input_text_list": MOCK_INPUTS,
                "reference_output_list": MOCK_REFERENCE_OUTPUTS,
            },
            MOCK_SUITE,
        ),
    ],
    ids=[
        "test-suite-default-columns",
        "test-suite-custom-columns",
        "test-suite-string-no-context",
    ],
)
def test_create_test_suite(params, expected, mock_client):
    suite = TestSuite(client=mock_client, **params)
    suite.client.get_suite_if_exists.assert_called_once_with(name=params["name"])
    suite.client.create_test_suite.assert_called_once()
    _, args, _ = suite.client.create_test_suite.mock_calls[0]
    assert_test_suite_equal(args[0], expected, check_page=False)


@pytest.mark.parametrize(
    "params,found_model",
    [
        (
            {"name": "test_suite", "scoring_method": "bertscore"},
            MOCK_SUITE_RESPONSE_WITH_PAGES,
        ),
        (
            {"name": "test_suite_custom", "scoring_method": CustomScorer()},
            MOCK_SUITE_CUSTOM_RESPONSE_WITH_PAGES,
        ),
    ],
    ids=["valid_builtin", "valid_custom"],
)
def test_reload_test_suite(params, found_model):
    mock_client = get_mock_client(suite_exists=True, mock_suite=found_model)
    suite = TestSuite(client=mock_client, **params)
    suite.client.get_suite_if_exists.assert_called_once_with(name=params["name"])
    suite.client.create_test_suite.assert_not_called()
    assert_test_suite_equal(suite._data, found_model)
    assert suite.scorer is not None


@pytest.mark.parametrize("candidate_column", ["custom_candidate", None])
def test_run_test_suite(candidate_column, test_suite_default):
    if candidate_column:
        run = test_suite_default.run(
            run_name="test_run",
            candidate_data=pd.DataFrame({candidate_column: MOCK_OUTPUTS}),
            candidate_column=candidate_column,
            model_name="my_very_special_gpt",
            save=False,
        )
    else:
        run = test_suite_default.run(
            run_name="test_run",
            candidate_data=pd.DataFrame({"candidate_output": MOCK_OUTPUTS}),
            model_name="my_very_special_gpt",
            save=False,
        )

    assert_test_run_equal(run, MOCK_RUN)


@pytest.mark.parametrize(
    "params",
    [
        {},
        {
            "input_column": "missing_prompt",
            "reference_data": pd.DataFrame({"prompt": [], "reference": []}),
        },
        {
            "reference_column": "missing_reference",
            "reference_data": pd.DataFrame({"prompt": [], "reference": []}),
        },
    ],
    ids=["missing-req-params", "input-missing", "reference-missing"],
)
def test_create_test_suite_invalid(params, mock_client):
    params.update({"name": "test_suite", "scoring_method": "bertscore"})
    with pytest.raises(ValueError):
        _ = TestSuite(client=mock_client, **params)


@pytest.mark.parametrize(
    "params",
    [
        {
            "candidate_column": "missing_candidate",
            "candidate_data": pd.DataFrame(
                {"candidate_output": ["three", "outputs", "provided"]}
            ),
        },
        {"candidate_data": pd.DataFrame({"candidate_output": ["only one output"]})},
    ],
    ids=["missing-candidate", "invalid-length"],
)
def test_run_test_suite_invalid(params, test_suite_default):
    params.update({"run_name": "test_run"})
    with pytest.raises(ValueError):
        _ = test_suite_default.run(**params)



from textstat import lexicon_count
from typing import List, Optional

from arthur_bench.exceptions import UserTypeError
from arthur_bench.scoring import Scorer


class WordCountMatch(Scorer):
    """
    Calculates how similar the number of words in the candidate output is to the the
    number of words in the reference output. Scores span from 0 to 1.
    A score of 1.0 indicates that there are the same number of words in the candidate
    output as in the reference output. Scores less than 1.0 are calculated as
    ((len_reference-delta)/len_reference) where delta is the absolute difference in
    word lengths between the candidate and reference outputs.
    All negative computed values are truncated to 0.
    Utilizes lexicon count, removing punctuations: https://pypi.org/project/textstat/

    """

    @staticmethod
    def name() -> str:
        return "word_count_match"

    def run_batch(
        self,
        candidate_batch: List[str],
        reference_batch: Optional[List[str]] = None,
        input_text_batch: Optional[List[str]] = None,
        context_batch: Optional[List[str]] = None,
    ) -> List[float]:
        if reference_batch is None:
            raise UserTypeError(
                "Reference Outputs must be provided for Word Count Match scorer. "
                "Please provide reference outputs to the test suite"
            )
        word_count_match = []
        for i in range(len(reference_batch)):
            len_ref = lexicon_count(reference_batch[i], removepunct=True)
            len_cand = lexicon_count(candidate_batch[i], removepunct=True)
            delta = abs(len_ref - len_cand)
            if delta > len_ref:
                word_count_match.append(0.0)
            else:
                word_count_match.append((len_ref - delta) / len_ref)

        return word_count_match



# flake8: noqa
from amplitude import Amplitude, BaseEvent
import uuid
import os
from enum import Enum
from arthur_bench.telemetry.config import TelemetryConfig
from arthur_bench.logger.logger import logger


AMPLITUDE_API_KEY = "bcc009e35e0b79daf089211108eed1de"
amplitude = Amplitude(AMPLITUDE_API_KEY)


class Telemetry(Enum):
    ON = 1
    OFF = 2
    LOG = 3


TRACK_USAGE_DATA: Telemetry


def set_track_usage_data(cfg: TelemetryConfig):
    global TRACK_USAGE_DATA
    TRACK_USAGE_DATA = Telemetry.ON if cfg.push_usage_data else Telemetry.OFF

    telemetry = os.getenv("BENCH_TELEMETRY_DISABLED", "0")
    if telemetry == "log":
        TRACK_USAGE_DATA = Telemetry.LOG
    elif telemetry != "0":
        TRACK_USAGE_DATA = Telemetry.OFF

    if TRACK_USAGE_DATA == Telemetry.ON and cfg.log_notice_of_usage_data:
        logger.warn(
            "Anonymous usage data is being collected by Arthur! For more details please see https://github.com/arthur-ai/bench/blob/develop/docs/source/telemetry.md."
        )


# A no-op if user opts out of data collection.
def send_event(event, user_id: uuid.UUID):
    if TRACK_USAGE_DATA == Telemetry.OFF:
        return

    elif TRACK_USAGE_DATA == Telemetry.LOG:
        logger.info(BaseEvent(user_id=str(user_id), **event))
        return

    logger.debug("Pushing usage data to Arthur.")
    amplitude.track(BaseEvent(user_id=str(user_id), **event))

    # Flush the event buffer
    amplitude.flush()



from textstat import flesch_reading_ease
from typing import List, Optional
from arthur_bench.scoring import Scorer


class Readability(Scorer):
    """
    Flesch Reading Ease Score: the higher the score, the easier to read.
    Scores of 100-90 correlate to a 5th grade reading level, while scores <10 are
    classified as "Extremely difficult to read, and best understood by university
    graduates."

    https://en.wikipedia.org/wiki/Flesch%E2%80%93Kincaid_readability_tests
    """

    @staticmethod
    def name() -> str:
        return "readability"

    @staticmethod
    def requires_reference() -> bool:
        return False

    def run_batch(
        self,
        candidate_batch: List[str],
        reference_batch: Optional[List[str]] = None,
        input_text_batch: Optional[List[str]] = None,
        context_batch: Optional[List[str]] = None,
    ) -> List[float]:
        return [flesch_reading_ease(i) for i in candidate_batch]



import pytest
from fastapi.testclient import TestClient

from arthur_bench.server.run_server import app
from tests.helpers import get_mock_client
from tests.fixtures.mock_responses import (
    MOCK_SUITES_JSON,
    MOCK_SUITE_RESPONSE_JSON,
    MOCK_RUNS_RESPONSE_JSON,
    MOCK_SUMMARY_RESPONSE_JSON,
    MOCK_RUN_RESPONSE_JSON,
)


@pytest.fixture
def mock_client():
    mock_arthur_client = get_mock_client(suite_exists=True)
    app.state.client = mock_arthur_client
    app.state.development = True  # disable telemetry during tests
    return TestClient(app)


TEST_SUITE_ID = "8b7ba080-8d14-42d2-9250-ec0edb96abd7"
RUN_ID = "af8466a8-6425-4ea5-85cb-ed952b26fa6c"


def test_get_test_suites(mock_client):
    resp = mock_client.get("/api/v3/bench/test_suites")
    assert resp.status_code == 200
    assert resp.json() == MOCK_SUITES_JSON


def test_get_test_suites_with_filters(mock_client):
    resp = mock_client.get(
        "/api/v3/bench/test_suites?scoring_method=bertscore&scoring_method=custom"
    )
    assert resp.status_code == 200
    assert resp.json() == MOCK_SUITES_JSON
    mock_client.app.state.client.get_test_suites.assert_called_once_with(
        page=1,
        page_size=5,
        sort=None,
        scoring_method=["bertscore", "custom"],
        name=None,
    )


def test_get_test_suite_by_id(mock_client):
    resp = mock_client.get(f"/api/v3/bench/test_suites/{TEST_SUITE_ID}")
    assert resp.status_code == 200
    assert resp.json() == MOCK_SUITE_RESPONSE_JSON


def test_get_test_runs(mock_client):
    resp = mock_client.get(f"/api/v3/bench/test_suites/{TEST_SUITE_ID}/runs")
    assert resp.status_code == 200
    assert resp.json() == MOCK_RUNS_RESPONSE_JSON


def test_get_suite_summary(mock_client):
    resp = mock_client.get(f"/api/v3/bench/test_suites/{TEST_SUITE_ID}/runs/summary")
    assert resp.status_code == 200
    assert resp.json() == MOCK_SUMMARY_RESPONSE_JSON


def test_get_run(mock_client):
    resp = mock_client.get(f"/api/v3/bench/test_suites/{TEST_SUITE_ID}/runs/{RUN_ID}")
    assert resp.status_code == 200
    assert resp.json() == MOCK_RUN_RESPONSE_JSON



from typing import List, Optional
from arthur_bench.scoring import Scorer
from arthur_bench.client.rest.client import ArthurClient
from arthur_bench.exceptions import ArthurUserError
from arthur_bench.models.scoring import HallucinationScoreRequest


class Hallucination(Scorer):
    """
    Score each output against a context using Arthur's hosted hallucination checker
    A score of 1.0 means the hallucination checker estimates the output is supported by
    the context
    A score of 0.0 means the hallucination checker found information in the output
    unsupported by the context
    """

    def __init__(self):
        self.client = ArthurClient()

    @staticmethod
    def name() -> str:
        return "hallucination"

    @staticmethod
    def requires_reference() -> bool:
        return False

    def to_dict(self, warn=False):
        return {}

    def run_batch(
        self,
        candidate_batch: List[str],
        reference_batch: Optional[List[str]] = None,
        input_text_batch: Optional[List[str]] = None,
        context_batch: Optional[List[str]] = None,
    ) -> List[float]:
        if context_batch is None:
            raise ArthurUserError("context is required for hallucination scoring")

        res = []
        for i in range(len(candidate_batch)):
            request = HallucinationScoreRequest(
                response=candidate_batch[i], context=context_batch[i]
            )
            response = self.client.bench.score_hallucination(request)
            score = float(response["hallucination"])
            res.append(score)
        return res



import pytest
import pandas as pd
from tests.helpers import FIXTURE_FILE_DIR
from tests.fixtures.mock_responses import (
    MOCK_SUITE_RESPONSE,
    MOCK_SUITE_JSON,
    MOCK_SUITE_WITH_NULL,
    MOCK_SUITE_WITH_NULL_JSON,
    MOCK_SUITE_WITH_SCORING_CONFIG,
    MOCK_SUITE_WITH_SCORING_JSON,
)
from arthur_bench.utils.loaders import (
    load_suite_from_json,
    load_suite_from_dataframe,
    load_suite_from_list,
    load_suite_from_csv,
)


@pytest.mark.parametrize(
    "filepath, expected",
    [
        ("mock_suite.json", "mock_suite_request"),
        ("mock_suite_without_optional.json", "mock_suite_request_optional"),
        ("mock_suite_with_typed_score.json", "mock_suite_request_optional"),
        ("mock_suite_mixed_null_refs.json", ValueError),
    ],
)
def test_load_suite_from_json(filepath, expected, request):
    if isinstance(expected, str):
        assert load_suite_from_json(
            FIXTURE_FILE_DIR / filepath
        ) == request.getfixturevalue(expected)
    elif issubclass(expected, Exception):
        with pytest.raises(expected):
            load_suite_from_json(FIXTURE_FILE_DIR / filepath)
    else:
        raise RuntimeError("bad test structure")


def test_load_suite_from_csv(mock_suite_cases):
    assert (
        load_suite_from_csv(
            FIXTURE_FILE_DIR / "mock_suite.csv", "input", "reference_output"
        )
        == mock_suite_cases
    )


def test_load_suite_from_list(mock_suite_cases):
    inputs = ["this is test input to a language model", "this is another test prompt"]
    outputs = ["this is test output from a language model", "this is a test response"]
    assert load_suite_from_list(inputs, outputs) == mock_suite_cases


def test_load_suite_from_df(mock_suite_cases):
    df = pd.DataFrame(
        {
            "input": [
                "this is test input to a language model",
                "this is another test prompt",
            ],
            "reference_output": [
                "this is test output from a language model",
                "this is a test response",
            ],
        }
    )
    assert (
        load_suite_from_dataframe(df, "input", "reference_output") == mock_suite_cases
    )


@pytest.mark.parametrize(
    "object, expected",
    [
        (MOCK_SUITE_RESPONSE, MOCK_SUITE_JSON),
        (MOCK_SUITE_WITH_NULL, MOCK_SUITE_WITH_NULL_JSON),
        (MOCK_SUITE_WITH_SCORING_CONFIG, MOCK_SUITE_WITH_SCORING_JSON),
    ],
    ids=["base_suite", "suite_with_null", "suite_with_custom_scoring"],
)
def test_suite_serialization(object, expected):
    assert object.json() == expected



import logging


class suppress_warnings:
    """
    A context-manager class to temporarily set the logging level for a logger to ERROR
    before returning it to its previous state.
    """

    def __init__(self, logger_name: str):
        self.logger = logging.getLogger(logger_name)
        self.original_state = (
            self.logger.level
        )  # note: we get actual level not effective level, could be UNSET

    def __enter__(self):
        self.logger.setLevel(logging.ERROR)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.logger.setLevel(
            self.original_state
        )  # it's ok to use setLevel() to set to anything, even NOTSET



# flake8: noqa

from langchain.prompts.chat import (
    ChatPromptTemplate,
    SystemMessagePromptTemplate,
    AIMessagePromptTemplate,
    HumanMessagePromptTemplate,
)

system_message_prompt = SystemMessagePromptTemplate.from_template(
    """Given the following context (extracted parts of a long document), a question, and an answer, decide if the final answer is correct.
You respond with a decision, either a 0, 1, or NA ONLY.
(0 = answer is incorrect, 1 = answer is correct, NA = you do not know. Don't try to make up an answer.).
If the question is not discussed in the provided context and the Answer says that the question is not discussed in the provided context, mark it as correct.
A correct answer is in accordance with fact and truth, answers the question posed, and is supported by the evidence found in the extracted parts of the context documents.
An incorrect answer has information that is conflicting or irrelevant to the extracted parts of the context documents, or has typos of words in the text,
or is a factually incorrect response to the questions."""
)
example_summaries_1 = HumanMessagePromptTemplate.from_template(
    """QUESTION: Compounds that are capable of accepting electrons, such as o 2 or f2, are called what?
========
CONTEXT: Oxidants and Reductants Compounds that are capable of accepting electrons, such as O 2 or F2, are called oxidants (or oxidizing agents) because they can oxidize other compounds. 
In the process of accepting electrons, an oxidant is reduced. Compounds that are capable of donating electrons, such as sodium metal or cyclohexane (C6H12), are called 
reductants (or reducing agents) because they can cause the reduction of another compound. In the process of donating electrons, a reductant is oxidized. These relationships are 
summarized in Equation 3.30: Equation 3.30 Saylor URL: http://www. saylor. org/books.
========
ANSWER: oxidants
DECISION:"""
)
example_choice_1 = AIMessagePromptTemplate.from_template("1")
example_summaries_2 = HumanMessagePromptTemplate.from_template(
    """QUESTION: What term in biotechnology means a genetically exact copy of an organism?
========
CONTEXT: But transgenic animals just have one novel gene. What about an animal with a whole new genome? Could a clone , a genetically exact copy of an organism, be 
developed using techniques associated with biotechnology? It could be argued that human cloning is one of the inevitable 
outcomes of modern biotechnology. It "simply" involves the removal of the nucleus from a somatic cell and its placement into an unfertilized egg cell whose nucleus has 
either been deactivated or removed. This new cell would mimic the zygote, the first diploid cell of a new organism. This new zygote is allowed to become established, 
and a few days later is placed into the uterus of a surrogate mother. Theoretically this would result in an individual genetically identical to the donor. Obviously, 
there are many ethical and legal issues associated with human cloning, and of course, it is not a "simple" procedure. But animal cloning is arguably a different story.
========
ANSWER: genome
DECISION:"""
)
example_choice_2 = AIMessagePromptTemplate.from_template("0")
example_summaries_3 = HumanMessagePromptTemplate.from_template(
    """QUESTION: What is the a mountain range in California?
========
CONTEXT: As you know, the surface of Earth is not flat. Some places are high, and some places are low. For example, mountain ranges like the Sierra Nevada in California 
or the Andes in South America are high above the surrounding areas. An accurate location must take into account the third dimension. Elevation is the height above or 
below sea level. Sea level refers to the height of the ocean’s surface. This is the midpoint between high and low tide. Sea level can vary from place to place, but scientists 
base their elevation measurements on the average, or mean, sea level to make sure they have a standard reference point.
========
ANSWER: Adirondacks
DECISION:"""
)
example_choice_3 = AIMessagePromptTemplate.from_template("0")
comparison_template = HumanMessagePromptTemplate.from_template(
    """Given the following context (extracted parts of a long document), a question, and an answer, decide if the final answer is correct.
You respond with a decision, either a 0, 1, or NA ONLY.
(0 = answer is incorrect, 1 = answer is correct, NA = you do not know. Don't try to make up an answer.).
If the question is not discussed in the provided context and the Answer says that the question is not discussed in the provided context, mark it as correct.
A correct answer is in accordance with fact and truth, and is supported by the evidence found in the extracted parts of the context documents.
An incorrect answer has information that is conflicting or irrelevant to the extracted parts of the context documents, or has typos of words in the text, or is a factually incorrect response to the questions.
QUESTION: {question}
========
CONTEXT: {context}
========
ANSWER: {answer}
DECISION:"""
)

DECIDE = ChatPromptTemplate.from_messages(
    [
        system_message_prompt,
        example_summaries_1,
        example_choice_1,
        example_summaries_2,
        example_choice_2,
        example_summaries_3,
        example_choice_3,
        comparison_template,
    ]
)



import pandas as pd
from typing import Optional, List, Tuple, Union

from arthur_bench.models.models import (
    TestCaseRequest,
)
from arthur_bench.exceptions import UserValueError
from arthur_bench.scoring import Scorer, scorer_from_string

from arthur_bench.utils.loaders import (
    load_suite_from_csv,
    load_suite_from_dataframe,
    load_suite_from_list,
    _validate_dataframe,
    get_file_extension,
)


def _initialize_scorer(
    scoring_method_arg: Union[str, Scorer], config: Optional[dict] = None
) -> Scorer:
    if isinstance(scoring_method_arg, str):
        scorer = scorer_from_string(scoring_method_arg)
        if config is not None:
            return scorer.from_dict(config)
        return scorer()
    else:
        return scoring_method_arg


def _load_suite_from_args(
    reference_data: Optional[pd.DataFrame] = None,
    reference_data_path: Optional[str] = None,
    input_column: str = "input",
    reference_column: Optional[str] = None,
    input_text_list: Optional[List[str]] = None,
    reference_output_list: Optional[List[str]] = None,
    requires_reference: bool = True,
) -> List[TestCaseRequest]:
    if requires_reference:
        if reference_column is None and reference_output_list is None:
            raise UserValueError(
                "scoring method requires reference data but no reference data was "
                "provided"
            )
    else:
        reference_column = None

    if reference_data is not None:
        return load_suite_from_dataframe(reference_data, input_column, reference_column)
    elif reference_data_path is not None:
        return load_suite_from_csv(reference_data_path, input_column, reference_column)
    elif input_text_list is not None:
        return load_suite_from_list(input_text_list, reference_output_list)
    else:
        raise UserValueError(
            "must specify data using either reference_data data frame, "
            "reference_data_path csv or input_text_list and reference_output_list "
            "strings"
        )


def _load_run_data_from_args(
    candidate_data: Optional[pd.DataFrame] = None,
    candidate_data_path: Optional[str] = None,
    candidate_column: str = "candidate_output",
    candidate_output_list: Optional[List[str]] = None,
    context_column: Optional[str] = None,
    context_list: Optional[List[str]] = None,
) -> Tuple[List[str], Optional[List[str]]]:
    if candidate_data is None and candidate_data_path is not None:
        if get_file_extension(candidate_data_path) != ".csv":
            raise UserValueError("filepath must be a csv file")
        candidate_data = pd.read_csv(candidate_data_path)

    if candidate_data is not None:
        _validate_dataframe(candidate_data, candidate_column)
        if context_column is not None:
            _validate_dataframe(candidate_data, context_column)
            return (
                candidate_data[candidate_column].values.tolist(),
                candidate_data[context_column].values.tolist(),
            )
        return candidate_data[candidate_column].values.tolist(), None

    elif candidate_output_list is not None:
        return candidate_output_list, context_list

    else:
        raise UserValueError(
            "must specify candidate data for run using candidate_data data frame, "
            "candidate_data_path csv, or candidate_output_list strings"
        )



# generated by datamodel-codegen:
#   filename:  admin.yaml
#   timestamp: 2023-06-21T19:01:59+00:00

from __future__ import annotations

from datetime import datetime
from typing import List, Optional

from pydantic import BaseModel


class User(BaseModel):
    id: str
    """
    the unique id of the user
    """
    first_name: Optional[str] = None
    """
    The user's first name
    """
    last_name: Optional[str] = None
    """
    The user's last name
    """
    email: str
    """
    The user's email
    """
    username: Optional[str] = None
    """
    The username the user can use to login
    """
    roles: Optional[List[str]] = None
    """
    The user's roles in the current organization.
    """
    alert_notifications_enabled: Optional[bool] = None
    """
    Whether or not the user will receive email notifications when alerts are triggered,
    defaults to 'false'
    """
    show_intro_sequence: Optional[bool] = None
    """
    Used by the Arthur dashboard to determine whether or not to show the user an intro
    sequence upon login
    """
    help_mode_enabled: Optional[bool] = None
    """
    Used by the Arthur dashboard to determine whether or not to show dashboard tooltips
    """
    created_at: Optional[datetime] = None
    """
    UTC timestamp of when the user was created
    """


class UserContext(BaseModel):
    name: Optional[str] = None
    """
    Name of the context.
    """
    id: Optional[str] = None
    """
    UUID of the context.
    """


class LoginRequest(BaseModel):
    login: str
    """
    either an email or a username
    """
    password: str


class AuthenticationInfo(BaseModel):
    organization_ids: List[str]
    """
    A list of organization IDs.
    """
    issuer: str
    """
    The identifier of the IDP managing this user.
    """
    external_user_id: Optional[str] = None
    """
    An identifier for an external-IdP token bearer. Populated if this user's token came
    from an IDP and the IDP configuration specified an oidc.CustomClaimNames that mapped
    UserID to a claim.
    """
    internal_user_id: Optional[str] = None
    """
    An identifier for an Arthur-internal user. Populated for Arthur-authenticated users
    with user tokens.
    """
    service_account_id: Optional[str] = None
    """
    An identifier for an Arthur service account. Populated for Arthur-authenticated
    service account tokens.
    """
    username: str
    """
    Either the Arthur username or the username specified by an external IDP. This will
    be set to arthur.ServiceAccountName for service account tokens.
    """
    first_name: str
    """
    The first name of the Arthur authenticated user, or the first name claim if the
    external IDP is configured with one.
    """
    last_name: str
    """
    The last name of the Arthur authenticated user, or the last name claim if the
    external IDP is configured with one.
    """
    email: str
    """
    The email of the Arthur authenticated user or the email of the external IDP user if
     the IDP is configured with that claim. For Arthur service accounts, this will be
     empty.
    """
    roles: List[str]
    """
    The list of roles that this user has. For Arthur tokens, there will always be one
     role in this array; however, there can be more than one for external providers.
    """


class UserResponse(BaseModel):
    """
    Represents an application user, if the client is using a service token then only
      organization_id and roles will be populated in the object
    """

    id: Optional[str] = None
    """
    the unique id of the user
    """
    organization_id: str
    """
    The ID of the users current context
    """
    organization_name: Optional[str] = None
    """
    The name of the users current context
    """
    first_name: Optional[str] = None
    """
    The user's first name
    """
    last_name: Optional[str] = None
    """
    The user's last name
    """
    email: Optional[str] = None
    """
    The user's email
    """
    username: Optional[str] = None
    """
    The username the user can use to login
    """
    roles: Optional[List[str]]
    """
    The user's roles
    """
    alert_notifications_enabled: Optional[bool] = None
    """
    Whether or not the user will receive email notifications when alerts are triggered,
      defaults to 'false'
    """
    show_intro_sequence: Optional[bool] = None
    """
    used by the Arthur dashboard to determine whether the user should be shown the intro
      sequence upon login
    """
    help_mode_enabled: Optional[bool] = None
    """
    Used by the Arthur dashboard to determine whether or not to show dashboard tooltips
    """
    plan: Optional[str] = None
    """
    string representation of what plan the org of the returned user is associated with
      (ie. self-service or paidSaas)
    """
    created_at: Optional[datetime] = None
    """
    UTC timestamp of when the user was created
    """
    contexts: Optional[List[UserContext]] = None
    """
    Contexts that the user has permissions in.
    """



# generated by datamodel-codegen:
#   filename:  common.yaml
#   timestamp: 2023-06-21T19:01:59+00:00

from __future__ import annotations

from typing import Dict, Optional, Union

from pydantic import BaseModel, Extra, Field


class KeyValueObject(BaseModel):
    class Config:
        extra = Extra.allow

    __root__: Optional[Dict[str, Union[float, str, bool]]] = None


class Page(BaseModel):
    __root__: int


class PageSize(BaseModel):
    __root__: int = Field(..., example=100)


class Sort(BaseModel):
    __root__: str
    """
    Must be supplied in the format [column_name] to denote asc sort by this column OR
    -[column_name] to denote desc sort by this column
    """


class StatusCounts(BaseModel):
    """
    Counts of the results by status
    """

    failure: float
    """
    The number of failed inferences in the request
    """
    success: float
    """
    The number of successful inferences in the request
    """
    total: float
    """
    The total number of inferences in the request
    """


class TotalCount(BaseModel):
    __root__: int = Field(..., example=2)
    """
    The total number of results the query returned
    """


class TotalPages(BaseModel):
    __root__: int = Field(..., example=1)
    """
    The number of result pages that this query generated, the ceiling of the total
    number of results divided by the page size
    """



from typing import List, Optional

from arthur_bench.exceptions import UserTypeError
from arthur_bench.scoring import Scorer


class ExactMatch(Scorer):
    """
    Returns 1 if candidate matches reference, 0 if candidate does not match reference.
    """

    def __init__(self, case_sensitive=True):
        self.case_sensitive = case_sensitive

    @staticmethod
    def name() -> str:
        return "exact_match"

    def run_batch(
        self,
        candidate_batch: List[str],
        reference_batch: Optional[List[str]] = None,
        input_text_batch: Optional[List[str]] = None,
        context_batch: Optional[List[str]] = None,
    ) -> List[float]:
        if reference_batch is None:
            raise UserTypeError(
                "Reference Outputs must be provided for Exact Match scorer. Please "
                "provide reference outputs to the test suite"
            )
        if not self.case_sensitive:
            return [
                float(reference_batch[i].lower() == candidate_batch[i].lower())
                for i in range(len(reference_batch))
            ]
        return [
            float(reference_batch[i] == candidate_batch[i])
            for i in range(len(reference_batch))
        ]



# flake8: noqa
from evaluate import load
import os

os.environ[
    "HF_ALLOW_CODE_EVAL"
] = "1"  # required for executing code using the HuggingFace code_eval metric
from typing import List, Optional
from tqdm import tqdm

from arthur_bench.exceptions import UserValueError
from arthur_bench.scoring import Scorer
from arthur_bench.scoring.scorer import SINGLE_ITEM_BATCH_DEFAULT


class PythonUnitTesting(Scorer):
    """
    Wrapping the HuggingFace code_eval metric

    Scores each candidate_output as a function against a pre-prepared unit test

    Note: considers any code with non-standard python libraries (e.g. numpy) to have an
    error

    https://huggingface.co/spaces/evaluate-metric/code_eval
    """

    def __init__(
        self,
        unit_test_dir: Optional[str] = None,
        unit_tests: Optional[List[str]] = None,
    ):
        self.evaluator = load("code_eval")  # type: ignore
        if unit_test_dir is not None:
            try:
                self.unit_tests = [
                    open(f"{unit_test_dir}/{x}").read()
                    for x in sorted(os.listdir(unit_test_dir))
                ]
            except FileNotFoundError as e:
                raise UserValueError(
                    f"Unable to read unit test files from unit_test_dir "
                    f"{unit_test_dir}: "
                ) from e
        else:
            if unit_tests is None:
                raise ValueError(
                    "Must create a PythonUnitTesting scoring method with either a "
                    "unit_test_dir parameter (to read unit tests from a file)"
                    " or a unit_tests parameter directly (to read unit tests as a list "
                    "of strings)"
                )
            self.unit_tests = unit_tests

    @staticmethod
    def name() -> str:
        return "python_unit_testing"

    @staticmethod
    def requires_reference() -> bool:
        return False

    def to_dict(self):
        return {"unit_tests": self.unit_tests}

    def run(
        self,
        candidate_outputs: List[str],
        reference_outputs: Optional[List[str]] = None,
        inputs: Optional[List[str]] = None,
        contexts: Optional[List[str]] = None,
        batch_size: int = SINGLE_ITEM_BATCH_DEFAULT,
    ) -> List[float]:
        res = []
        for i in tqdm(range(len(candidate_outputs))):
            pass_at_k, _ = self.evaluator.compute(
                references=[self.unit_tests[i]], predictions=[[candidate_outputs[i]]]
            )
            res.append(pass_at_k["pass@1"])
        return res

    def run_batch(
        self,
        candidate_batch: List[str],
        reference_batch: Optional[List[str]] = None,
        input_text_batch: Optional[List[str]] = None,
        context_batch: Optional[List[str]] = None,
    ) -> List[float]:
        raise NotImplementedError(
            "run_batch is not implemented for this scorer. "
            "Use PythonUnitTesting.run(candidates) instead."
        )



# flake8: noqa
from .exceptions import (
    ArthurError,
    ArthurUserError,
    ArthurInternalError,
    MissingParameterError,
    UserValueError,
    UserTypeError,
    MethodNotApplicableError,
    ResponseClientError,
    UnauthorizedError,
    PaymentRequiredError,
    ForbiddenError,
    NotFoundError,
    ExpectedParameterNotFoundError,
    InternalValueError,
    InternalTypeError,
    ResponseServerError,
    ResponseRedirectError,
)



from enum import Enum
from typing import Dict

from .scorer import Scorer
from .bertscore import BERTScore
from .qa_quality import QAQualityCorrectness
from .summary_quality import SummaryQuality
from .exact_match import ExactMatch
from .hallucination import Hallucination
from .readability import Readability
from .word_count_match import WordCountMatch
from .specificity import Specificity
from .hedging_language import HedgingLanguage
from .python_unit_testing import PythonUnitTesting
from arthur_bench.exceptions import UserValueError


class ScoringMethodName(str, Enum):
    BERTScore = "bertscore"
    SummaryQuality = "summary_quality"
    QACorrectness = "qa_correctness"
    ExactMatch = "exact_match"
    Hallucination = "hallucination"
    Readability = "readability"
    WordCountMatch = "word_count_match"
    Specificity = "specificity"
    HedgingLanguage = "hedging_language"
    PythonUnitTesting = "python_unit_testing"


SCORING_METHOD_CLASS_MAP: Dict[str, type[Scorer]] = {
    ScoringMethodName.BERTScore: BERTScore,
    ScoringMethodName.QACorrectness: QAQualityCorrectness,
    ScoringMethodName.SummaryQuality: SummaryQuality,
    ScoringMethodName.ExactMatch: ExactMatch,
    ScoringMethodName.Hallucination: Hallucination,
    ScoringMethodName.Readability: Readability,
    ScoringMethodName.WordCountMatch: WordCountMatch,
    ScoringMethodName.Specificity: Specificity,
    ScoringMethodName.HedgingLanguage: HedgingLanguage,
    ScoringMethodName.PythonUnitTesting: PythonUnitTesting,
}


def scorer_from_string(method: str) -> type[Scorer]:
    if method in SCORING_METHOD_CLASS_MAP:
        return SCORING_METHOD_CLASS_MAP[method]
    else:
        raise UserValueError(f"Unknown scorer string {method}")



from textstat import lexicon_count
from typing import List, Optional
from arthur_bench.scoring import Scorer
from wordfreq import word_frequency
from collections import Counter
import nltk
import re


class Specificity(Scorer):
    """
    Returns a score from 0.0 to 1.0 indicating how specific the candidate output
    language is. Higher scores indicate that the language is more specific,
    Lower scores indicate more vague language.

    Specificity is computed through detecting words that indicate vagueness (predefined)
    determing how rare the words used are according to word frequencies calculated by
    popular nlp corpora, and detecting use of proper nouns and numbers.
    """

    @staticmethod
    def name() -> str:
        return "specificity"

    @staticmethod
    def requires_reference() -> bool:
        return False

    def __init__(self):
        nltk.download("punkt")
        nltk.download("averaged_perceptron_tagger")

    def get_num_vague_words(self, candidate_output: str) -> int:
        """
        Returns number of words in candidate_output which are is a list of pre-defined
        vague words.
        """
        vague_word_list = [
            "around",
            "about",
            "almost",
            "basically",
            "approximately",
            "nearly",
            "roughly",
            "some",
            "most",
            "least",
            "good",
            "bad",
            "nice",
            "great",
            "may",
            "could",
            "certain",
            "most",
            "might",
            "usually",
            "typically",
            "like",
            "sure",
        ]
        counter = 0
        candidate_output = candidate_output.lower()
        for v in vague_word_list:
            i = candidate_output.count(v)
            counter += i
        len_s = lexicon_count(candidate_output, removepunct=True)

        non_vague_prop = (
            len_s - counter
        ) / len_s  # proportion of words which are not vague
        non_vague_prop = (non_vague_prop / 0.2) - 4.0
        non_vague_prop = max(0.0, non_vague_prop)  # normalize
        return non_vague_prop

    def get_mean_word_freq(self, candidate_output: str) -> float:
        """
        Returns mean word frequency of candidate output. Higher values indicate that
        moree common words on average are used in the candidate output.
        Considers only words with frequency <0.001, truncating probability of words with
        higher frequencies to 0.001.
        """
        punct_regex = r"[^\w\s\'/]"
        # don't replace apostrophes
        text = re.sub(punct_regex, "", candidate_output)
        text = re.sub(r"[/]", " ", text)
        text_list = text.split()
        word_freqs = [word_frequency(word, "en") for word in text_list]

        # truncates all probability of words w freq >0.001 to 0.001
        filtered = [0.001 if i > 0.001 else i for i in word_freqs]
        if len(filtered) == 0:
            wf = 0.001
        else:
            wf = sum(filtered) / len(filtered)

        adj_freq = 2.0 - (wf / 0.0004)  # reverse scale and normalize
        adj_freq = max(0.0, adj_freq)
        adj_freq = min(1.0, adj_freq)

        return adj_freq

    def get_pn_and_num(self, candidate_output: str) -> int:
        """
        Returns total number of Proper Nouns and Numbers in candidate output.
        Determined heuristically via NNP and CD nltk tags.
        """
        # get number of proper nouns and numbers
        tokens = nltk.word_tokenize(candidate_output)
        tags = nltk.pos_tag(tokens)
        counts = Counter(tag for word, tag in tags)
        pn = counts["NNP"] + counts["CD"]
        len_s = lexicon_count(candidate_output, removepunct=True)
        pn_prop = 5.0 * (pn / len_s)
        pn_prop = min(1.0, pn_prop)
        return pn_prop

    def run_batch(
        self,
        candidate_batch: List[str],
        reference_batch: Optional[List[str]] = None,
        input_text_batch: Optional[List[str]] = None,
        context_batch: Optional[List[str]] = None,
    ) -> List[float]:
        res = []
        for i in range(len(candidate_batch)):
            c = candidate_batch[i]

            vague_prop = self.get_num_vague_words(c)
            adj_freq = self.get_mean_word_freq(c)
            pn_prop = self.get_pn_and_num(c)

            s = (0.33 * vague_prop) + (0.33 * adj_freq) + (0.33 * pn_prop)  # aggregate
            res.append(s)

        return res



import os
import json
import argparse
import logging
from pathlib import Path
import uuid
from typing import Optional, Annotated, List, Union

try:
    import uvicorn
    from fastapi import FastAPI, Request, HTTPException, Query
    from fastapi.staticfiles import StaticFiles
    from fastapi.middleware.cors import CORSMiddleware


except ImportError as e:
    raise ImportError(
        "Can't run Bench Server without server dependencies, to install run: "
        "pip install arthur-bench[server]"
    ) from e

from arthur_bench.client.local.client import LocalBenchClient
from arthur_bench.exceptions import NotFoundError
from arthur_bench.telemetry.telemetry import send_event, set_track_usage_data
from arthur_bench.telemetry.config import get_or_persist_id, persist_usage_data

logger = logging.getLogger(__name__)

app = FastAPI()
app.state.development = False

origins = ["http://localhost:8000", "http://127.0.0.1:8000"]

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

FRONT_END_DIRECTORY = Path(__file__).parent / "js" / "dist"


@app.get("/api/v3/bench/test_suites")
def test_suites(
    request: Request,
    page: int = 1,
    page_size: int = 5,
    sort: Optional[str] = None,
    scoring_method: Annotated[Union[List[str], None], Query()] = None,
    name: Optional[str] = None,
):
    client = request.app.state.client
    suite_resp = client.get_test_suites(
        page=page,
        page_size=page_size,
        sort=sort,
        scoring_method=scoring_method,
        name=name,
    )

    if not request.app.state.development:
        send_event(
            {
                "event_type": "test_suites_load",
                "event_properties": {
                    "num_test_suites_load": len(suite_resp.test_suites),
                    "test_suites_all": [
                        suite.scoring_method.name for suite in suite_resp.test_suites
                    ],
                },
            },
            request.app.state.user_id,
        )
    return suite_resp


@app.get("/api/v3/bench/test_suites/{test_suite_id}")
def test_suite(
    request: Request, test_suite_id: uuid.UUID, page: int = 1, page_size: int = 5
):
    client = request.app.state.client
    try:
        suite_resp = client.get_test_suite(
            test_suite_id=str(test_suite_id), page=page, page_size=page_size
        )
    except NotFoundError as e:
        return HTTPException(status_code=404, detail=str(e))
    return suite_resp


@app.get("/api/v3/bench/test_suites/{test_suite_id}/runs")
def test_runs(
    request: Request,
    test_suite_id: uuid.UUID,
    page: int = 1,
    page_size: int = 5,
    sort: Optional[str] = None,
):
    client = request.app.state.client
    try:
        run_resp = client.get_runs_for_test_suite(
            test_suite_id=str(test_suite_id), page=page, page_size=page_size, sort=sort
        )
    except NotFoundError as e:
        return HTTPException(status_code=404, detail=str(e))
    suite_resp = client.get_test_suite(test_suite_id=str(test_suite_id))

    if not request.app.state.development:
        send_event(
            {
                "event_type": "test_runs_load",
                "event_properties": {
                    "test_runs_all": [
                        str(run.created_at) for run in run_resp.test_runs
                    ],
                    "scoring_method_real": suite_resp.scoring_method.name,
                },
            },
            request.app.state.user_id,
        )
    return run_resp


@app.get("/api/v3/bench/test_suites/{test_suite_id}/runs/summary")
def test_suite_summary(
    request: Request,
    test_suite_id: uuid.UUID,
    page: int = 1,
    page_size: int = 5,
    run_id: Optional[uuid.UUID] = None,
):
    client = request.app.state.client
    try:
        summary_resp = client.get_summary_statistics(
            test_suite_id=str(test_suite_id),
            page=page,
            page_size=page_size,
            run_id=str(run_id) if run_id is not None else None,
        )
    except NotFoundError as e:
        return HTTPException(status_code=404, detail=str(e))
    return summary_resp


@app.get("/api/v3/bench/test_suites/{test_suite_id}/runs/{run_id}")
def test_run_results(
    request: Request,
    test_suite_id: uuid.UUID,
    run_id: uuid.UUID,
    page: int = 1,
    page_size: int = 5,
):
    client = request.app.state.client

    try:
        run_resp = client.get_test_run(
            test_suite_id=str(test_suite_id),
            test_run_id=str(run_id),
            page=page,
            page_size=page_size,
        ).json(by_alias=True)
    except NotFoundError as e:
        return HTTPException(status_code=404, detail=str(e))
    run = json.loads(run_resp)
    return run


if os.path.exists(FRONT_END_DIRECTORY):
    app.mount(
        "/", StaticFiles(directory=FRONT_END_DIRECTORY, html=True), name="frontend"
    )
else:
    logger.warning(
        "frontend files not found. if running package from source, "
        "frontend files must be built locally"
    )


def run():
    # parser needs to go in this function for compatibility with packaging
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--directory",
        required=False,
        help="optional directory override to run as root for bench server ",
    )

    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--enable_push_usage_data",
        help="Enable sending anonymous usage data",
        required=False,
        action="store_true",
    )
    group.add_argument(
        "--disable_push_usage_data",
        help="Disable sending anonymous usage data",
        required=False,
        action="store_true",
    )
    args = parser.parse_args()
    if args.enable_push_usage_data:
        persist_usage_data(True)
        return
    if args.disable_push_usage_data:
        persist_usage_data(False)
        return

    client = LocalBenchClient(args.directory)
    app.state.client = client

    config = get_or_persist_id()
    set_track_usage_data(config)
    app.state.user_id = config.user_id

    uvicorn.run(
        "arthur_bench.server.run_server:app",
        host="127.0.0.1",
        port=8000,
        log_level="info",
    )


if __name__ == "__main__":
    run()



from bert_score import BERTScorer
from typing import List, Optional
from arthur_bench.scoring import Scorer
from arthur_bench.scoring.utils import suppress_warnings

DEFAULT_MODEL = "microsoft/deberta-v3-base"

# weight precision and recall differently
# for now, we have observed in experiments that
# recall is much more correlated with human judgment than precision
PRECISION_WEIGHT = 0.1
RECALL_WEIGHT = 0.9


class BERTScore(Scorer):
    """
    Tailored bert score implementation.

    https://arxiv.org/abs/1904.09675
    """

    @staticmethod
    def name() -> str:
        return "bertscore"

    def __init__(self, model_type=DEFAULT_MODEL, precision_weight=PRECISION_WEIGHT):
        """
        Tailored bert score implementation.

        :param model_type: the underlying language model to extract embeddings from
        :param precision_weight: the weight given to the precision term in calculating
            bertscore
        """
        self.precision_weight = precision_weight
        self.recall_weight = 1 - precision_weight

        with suppress_warnings("transformers"):
            self.model = BERTScorer(lang="en", model_type=model_type)

    def to_dict(self, warn=False):
        return {
            "precision_weight": self.precision_weight,
            "model_type": self.model.model_type,
        }

    def run_batch(
        self,
        candidate_batch: List[str],
        reference_batch: Optional[List[str]] = None,
        input_text_batch: Optional[List[str]] = None,
        context_batch: Optional[List[str]] = None,
    ) -> List[float]:
        # get precision, recall, and F1 score from bert_score package
        p, r, _ = self.model.score(candidate_batch, reference_batch, verbose=False)

        # return a BERTScore using our weighting of precision and recall (instead of F1
        # which weights them equally)
        return (self.precision_weight * p + self.recall_weight * r).tolist()



import logging
import tiktoken
from typing import List, Optional, Tuple
from langchain.chains import LLMChain
from langchain.chat_models import ChatOpenAI

from arthur_bench.exceptions import UserValueError, UserTypeError

from arthur_bench.scoring import Scorer
from arthur_bench.scoring.scorer import SINGLE_ITEM_BATCH_DEFAULT
from arthur_bench.scoring.prompts.summary_quality import COMPARE


CONTEXT_WINDOW_MAP = {
    "gpt-3.5-turbo": 4096,
    "gpt-4": 8192,
    "gpt-3.5-turbo-16k": 16384,
    "gpt-4-32k": 32768,
}
EVALUATOR_MODEL = "gpt-3.5-turbo"
TIKTOKEN_ENCODER = tiktoken.get_encoding("cl100k_base")
TIKTOKEN_ERROR_PADDING = 150
LLM_CHOICE_OPTIONS = {"0": 0.0, "1": 1.0, "tie": 0.5}

logger = logging.getLogger(__name__)


def truncate_input_text(input_text, ref_output, cand_output) -> Tuple[str, bool]:
    """Truncates the input_text to fit in LLM evaluator context

    Truncate the input text so that the filled-in COMPARE prompt
    which contains {input text + summary A + summary B} fits in the evaluator context
    window

    Returns the tuple (text, whether text was truncated)
    """
    llm_prompt_untruncated = COMPARE.format(
        text=input_text, summary_A=ref_output, summary_B=cand_output
    )
    input_text_tokens = TIKTOKEN_ENCODER.encode(input_text)
    llm_prompt_tokens = TIKTOKEN_ENCODER.encode(llm_prompt_untruncated)
    num_to_truncate_from_input_text_tokens = (
        len(llm_prompt_tokens)
        - CONTEXT_WINDOW_MAP[EVALUATOR_MODEL]
        + TIKTOKEN_ERROR_PADDING
    )
    truncated = False
    if num_to_truncate_from_input_text_tokens > 0:
        input_text_tokens_truncated = input_text_tokens[
            :-num_to_truncate_from_input_text_tokens
        ]
        input_text = TIKTOKEN_ENCODER.decode(input_text_tokens_truncated)
        truncated = True
    return input_text, truncated


class SummaryQuality(Scorer):
    """
    Comprehensive measure of summarization quality compared to a reference summary.
    """

    def __init__(self):
        self.evaluator = LLMChain(llm=ChatOpenAI(temperature=0), prompt=COMPARE)

    @staticmethod
    def name() -> str:
        return "summary_quality"

    def to_dict(self, warn=False):
        return {}

    def run(
        self,
        candidate_outputs: List[str],
        reference_outputs: Optional[List[str]] = None,
        inputs: Optional[List[str]] = None,
        contexts: Optional[List[str]] = None,
        batch_size: int = SINGLE_ITEM_BATCH_DEFAULT,
    ) -> list:
        if inputs is None:
            raise TypeError(
                "Inputs must be provided for Summary Quality scorer, got None"
            )
        if reference_outputs is None:
            raise TypeError(
                "Reference Outputs must be provided for Summary Quality scorer, "
                "got None"
            )
        # truncate inputs if needed
        truncated_inputs = []
        num_truncated = 0
        for i in range(len(inputs)):
            inp, truncated = truncate_input_text(
                inputs[i], reference_outputs[i], candidate_outputs[i]
            )
            num_truncated += int(truncated)

            # add to list we'll actually use
            truncated_inputs.append(inp)

        if num_truncated > 0:
            logger.warning(
                f"Truncated {num_truncated} out of {len(inputs)} total summary inputs "
                f"to {CONTEXT_WINDOW_MAP[EVALUATOR_MODEL]} characters"
            )

        return super().run(
            candidate_outputs, reference_outputs, truncated_inputs, contexts, batch_size
        )

    def run_batch(
        self,
        candidate_batch: List[str],
        reference_batch: Optional[List[str]] = None,
        input_text_batch: Optional[List[str]] = None,
        context_batch: Optional[List[str]] = None,
    ) -> List[float]:
        """
        Summary quality requires input_text_batch.
        """
        if input_text_batch is None:
            raise UserValueError(
                "input text is required for this scorer. Please provide a dataframe "
                "column or a list of your "
                "input text strings in the Test Suite."
            )
        if reference_batch is None:
            raise UserTypeError(
                "Reference Outputs must be provided for Summary Quality scorer. Please "
                "provide reference outputs to the test suite"
            )

        if context_batch is not None:
            raise UserValueError(
                "using context is not currently supported for summary quality"
            )

        res = []
        for i in range(len(input_text_batch)):
            # run LLMChain to choose whether summary A or summary B is a better summary
            # of the input text

            choice = self.evaluator(
                {
                    "text": input_text_batch[i],
                    "summary_A": reference_batch[i],
                    "summary_B": candidate_batch[i],
                }
            )

            # return -1.0 if the LLMChain returns an invalid result
            if "text" in choice:
                res.append(LLM_CHOICE_OPTIONS.get(choice["text"][:3], -1.0))
            else:
                res.append(-1.0)
        return res



import logging
import os
from distutils.util import strtobool
from getpass import getpass
from platform import system
from typing import Optional


from arthur_bench.client.auth.helpers import get_current_org, user_login
from arthur_bench.client.auth.refresh import AuthRefresher
from arthur_bench.client.http.requests import HTTPClient
from arthur_bench.exceptions import MissingParameterError, UserValueError

# import sub-clients
from arthur_bench.client.rest.admin.client import ArthurAdminClient
from arthur_bench.client.rest.bench.client import ArthurBenchClient

# end import sub-clients

from arthur_bench.version import __version__

UNKNOWN_ORG_ID = "unknown-org"
ORG_ID_HEADER = "Arthur-Organization-ID"


logger = logging.getLogger(__name__)


class ArthurClient:
    def __init__(
        self,
        url: Optional[str] = None,
        login: Optional[str] = None,
        password: Optional[str] = None,
        api_key: Optional[str] = None,
        organization_id: Optional[str] = None,
        verify_ssl: Optional[bool] = None,
        allow_insecure: bool = False,
        offline: bool = False,
    ):
        """
        REST client for logging and fetching data stored in the Arthur platform
        """
        # basic values
        if url is None:
            url = os.getenv("ARTHUR_API_URL")
        if url is None:
            raise MissingParameterError(
                "You must provide a URL through the 'url' parameter or the "
                "ARTHUR_API_URL environment variable"
            )
        if verify_ssl is None:  # if not provided
            # read from env var, or default to True if not present
            try:
                verify_ssl = bool(strtobool(os.getenv("ARTHUR_VERIFY_SSL", "true")))
            except ValueError:
                raise UserValueError(
                    f"ARTHUR_VERIFY_SSL environment variable must be a boolean value, "
                    f"got {os.getenv('ARTHUR_VERIFY_SSL')}"
                )

        # authorization
        if login is None:
            login = os.getenv("ARTHUR_LOGIN")
        if password is None:
            password = os.getenv("ARTHUR_PASSWORD")
        if api_key is None and login is None:
            api_key = os.getenv("ARTHUR_API_KEY")
        # validate only login or api key
        if login is not None and api_key is not None:
            raise UserValueError(
                "You may not provide both a login and api key, please ensure you are "
                "supplying only one through the login/api_key parameters and "
                "ARTHUR_LOGIN/ARTHUR_API_KEY environment "
                "variables"
            )
        if login is not None:  # login if provided
            # if password not supplied, get it from input
            if password is None:
                password = getpass(f"Please enter password for {login}: ")

            # Get session token from login and password
            auth_token = user_login(
                api_http_host=url, login=login, password=password, verify_ssl=verify_ssl
            )
            # create an auth refresher
            auth_refresher = AuthRefresher(
                url=url, login=login, password=password, verify_ssl=verify_ssl
            )
            header_refresh_func = auth_refresher.refresh
        elif api_key is not None:  # if api key provided, set that
            auth_token = api_key
            header_refresh_func = None
        else:
            raise MissingParameterError(
                "No authentication provided. Please supply a login (username or email) "
                "through the 'login' parameter or ARTHUR_LOGIN environment variable."
                "\n\n If this is a production environment, alternatively consider "
                "providing an API key through the 'api_key' parameter or ARTHUR_API_KEY"
                "environment variable."
            )

        # org id
        if offline:
            if organization_id is not None:
                raise UserValueError(
                    "You cannot specify an organization ID if you are offline."
                )
            else:
                organization_id = UNKNOWN_ORG_ID
        else:  # if online
            # fill org ID with environment variable if not provided
            if organization_id is None:
                organization_id = os.getenv("ARTHUR_ORGANIZATION_ID")
            # if still no org ID, fetch it from the API
            if organization_id is None:
                organization_id = get_current_org(
                    url, auth_token, verify_ssl=verify_ssl
                )

        # TODO: consider having the SDK override this?
        user_agent = f"arthur-client/{__version__} (system={system()})"
        headers = {
            "Accept": "application/json",
            "Authorization": auth_token,
            "User-Agent": user_agent,
            ORG_ID_HEADER: organization_id,
        }

        # setup http client construction arguments
        client_kwargs = {
            "base_url": url,
            "default_headers": headers,
            "verify_ssl": verify_ssl,
            "allow_insecure": allow_insecure,
            "header_refresh_func": header_refresh_func,
        }

        client = HTTPClient(**client_kwargs)  # type: ignore
        # create client object for each client subcomponent
        self.admin = ArthurAdminClient(client)
        self.bench = ArthurBenchClient(client)
        # end client object creation



from typing import Optional, TypeVar, List
from abc import ABC, abstractmethod

from arthur_bench.models.models import (
    PaginatedTestSuites,
    CreateRunResponse,
    CreateRunRequest,
    PaginatedRuns,
    PaginatedRun,
    TestSuiteRequest,
    PaginatedTestSuite,
    TestSuiteSummary,
)

from arthur_bench.exceptions import ArthurInternalError

TBenchClient = TypeVar("TBenchClient", bound="BenchClient")


class BenchClient(ABC):
    @abstractmethod
    def get_test_suites(
        self,
        name: Optional[str] = None,
        sort: Optional[str] = None,
        scoring_method: Optional[List[str]] = None,
        page: int = 1,
        page_size: int = 5,
    ) -> PaginatedTestSuites:
        raise NotImplementedError

    @abstractmethod
    def create_test_suite(self, json_body: TestSuiteRequest) -> PaginatedTestSuite:
        raise NotImplementedError

    @abstractmethod
    def get_test_suite(
        self, test_suite_id: str, page: int = 1, page_size: int = 5
    ) -> PaginatedTestSuite:
        raise NotImplementedError

    @abstractmethod
    def get_runs_for_test_suite(
        self,
        test_suite_id: str,
        sort: Optional[str] = None,
        page: int = 1,
        page_size: int = 5,
    ) -> PaginatedRuns:
        raise NotImplementedError

    @abstractmethod
    def create_new_test_run(
        self, test_suite_id: str, json_body: CreateRunRequest
    ) -> CreateRunResponse:
        raise NotImplementedError

    @abstractmethod
    def get_test_run(
        self,
        test_suite_id: str,
        test_run_id: str,
        page: int = 1,
        page_size: int = 5,
        sort: Optional[bool] = None,
    ) -> PaginatedRun:
        raise NotImplementedError

    @abstractmethod
    def get_summary_statistics(
        self,
        test_suite_id: str,
        run_id: Optional[str] = None,
        page: int = 1,
        page_size: int = 5,
    ) -> TestSuiteSummary:
        raise NotImplementedError

    @abstractmethod
    def delete_test_suite(self, test_suite_id: str):
        raise NotImplementedError

    @abstractmethod
    def delete_test_run(self, test_suite_id: str, test_run_id: str):
        raise NotImplementedError

    def get_suite_if_exists(self, name: str) -> Optional[PaginatedTestSuite]:
        """
        Get a full test suite with name if it exists.

        :param client: BenchClient object for fetching test suite data
        :return: complete test suite with all test cases joined,
            or None if no suite with name exists
        :raises ArthurInternalError: if using a client that does not support pagination
        """
        test_suite_resp = self.get_test_suites(name=name)
        if len(test_suite_resp.test_suites) > 0:
            # we enforce name validation, so there should ever only be one
            query_params = {"page_size": 100}

            suite = self.get_test_suite(
                str(test_suite_resp.test_suites[0].id), **query_params
            )

            current_page: int
            total_pages: int
            if suite.page is None or suite.total_pages is None:
                raise ArthurInternalError("expected paginated response")

            current_page = suite.page + 1
            total_pages = suite.total_pages
            while current_page <= total_pages:
                query_params["page"] = current_page
                suite_next_page = self.get_test_suite(
                    str(test_suite_resp.test_suites[0].id), **query_params
                )

                suite.test_cases.extend(suite_next_page.test_cases)

                if suite_next_page.page is None or suite_next_page.total_pages is None:
                    raise ArthurInternalError("expected paginated response")
                total_pages = suite_next_page.total_pages
                current_page = suite_next_page.page + 1
            return suite
        return None

    def check_run_exists(self, suite_id: str, run_name: str) -> bool:
        """
        Check if run with given name if it exists for suite with id suite_id

        :param client: BenchClient object for fetching test suite data
        :param suite_id: the id of the test suite to check run names
        :param run_name: the test run name to check for
        :return: True if run with name is found, False otherwise
        :raises ArthurInternalError: if using a client that does not support pagination
        """

        page_size = 100
        current_page = 1
        total_pages = 1
        page = 1

        while current_page <= total_pages:
            runs = self.get_runs_for_test_suite(
                suite_id, page=page, page_size=page_size
            )
            for run_metadata in runs.test_runs:
                if run_metadata.name == run_name:
                    return True
            if runs.page is None or runs.total_pages is None:
                raise ArthurInternalError("expected paginated response")
            total_pages = runs.total_pages
            current_page = runs.page + 1
        return False



import os
import pandas as pd
from pathlib import Path
from typing import Optional, List, Union

from arthur_bench.models.models import (
    TestCaseRequest,
    TestSuiteRequest,
)

from arthur_bench.exceptions import UserValueError


def get_file_extension(filepath: Union[str, os.PathLike]) -> str:
    _, ext = os.path.splitext(filepath)
    return ext


def _validate_dataframe(data: pd.DataFrame, column: str):
    if column not in data.columns:
        raise UserValueError(f"column {column} not found in dataset")


def load_suite_from_json(filepath: Union[str, os.PathLike]) -> TestSuiteRequest:
    """
    Load a full test suite from a json file.

    :param filepath: string or pathlike object pointing to json file containing test
        suite data
    """
    if get_file_extension(filepath) != ".json":
        raise UserValueError("filepath must be json file")
    if isinstance(filepath, str):
        filepath = Path(filepath)
    return TestSuiteRequest.parse_file(filepath)  # type: ignore


def load_suite_from_dataframe(
    data: pd.DataFrame, input_column: str, reference_column: Optional[str] = None
) -> List[TestCaseRequest]:
    """
    Load test case data from a pandas dataframe.

    :param data: dataframe where each row is a test case consisting of a column for i
        input and a column for reference
    :param input_column: column in dataframe containing inputs
    :param reference_column: column in dataframe containing reference outputs
    """
    _validate_dataframe(data, input_column)
    if reference_column is not None:
        _validate_dataframe(data, reference_column)
        suite = data.rename(
            columns={input_column: "input", reference_column: "reference_output"}
        )[["input", "reference_output"]]
    else:
        suite = data.rename(columns={input_column: "input"})[["input"]]
    if len(suite) < 1:
        raise UserValueError("test suite must have at least 1 test case")
    return [TestCaseRequest(**row) for row in suite.to_dict("records")]


def load_suite_from_csv(
    filepath: Union[str, os.PathLike],
    input_column: str,
    reference_column: Optional[str] = None,
) -> List[TestCaseRequest]:
    """
    Load test case data from csv file.

    :param filepath: string or pathlike object pointing to csv file
    :param input_column: column in file containing inputs
    :param reference_column: column in file containing reference outputs
    """
    if get_file_extension(filepath) != ".csv":
        raise UserValueError("filepath must be a csv file")
    return load_suite_from_dataframe(
        pd.read_csv(filepath), input_column, reference_column
    )


def load_suite_from_list(
    inputs: List[str], reference_outputs: Optional[List[str]]
) -> List[TestCaseRequest]:
    """
    Load test case data from lists of strings.

    :param inputs: list of string inputs for each test case
    :param reference_outputs: list of string reference outputs for each input
    """
    if len(inputs) < 1:
        raise UserValueError("test suite must have at least 1 test case")

    if reference_outputs is not None:
        if len(inputs) != len(reference_outputs):
            raise UserValueError("inputs and reference_outputs must be the same length")
        return [
            TestCaseRequest(input=i, reference_output=o)
            for i, o in zip(inputs, reference_outputs)
        ]

    return [TestCaseRequest(input=i, reference_output=None) for i in inputs]



import requests
from typing import Dict, Tuple, cast
from http import HTTPStatus
from requests.cookies import RequestsCookieJar

# import http client
from arthur_bench.client.http.requests import HTTPClient

from arthur_bench.client.rest.admin.models import (
    LoginRequest,
    AuthenticationInfo,
    UserResponse,
    User,
)


PATH_PREFIX = "/api/v3"


class ArthurAdminClient:
    """
    A Python client to interact with the Arthur Admin API
    """

    def __init__(self, http_client: HTTPClient):
        """
        Create a new ArthurAdminClient from an HTTPClient

        :param http_client: the :class:`~arthurai.client.http.requests.HTTPClient` to
        use for underlying requests
        """
        self.http_client = http_client
        self.http_client.set_path_prefix(PATH_PREFIX)

    def login(self, json_body: LoginRequest) -> Tuple[User, RequestsCookieJar]:
        """
        If the login attempt is successful, the user will be returned in the response
        body and an HttpOnly, set-cookie \"Authorization\" header will be returned
        that contains a JWT to be used in subsequent requests to the API in either
        the \"Authorization\" or cookie header

        :param json_body:
        """

        raw_resp = cast(
            requests.Response,
            self.http_client.post(
                "/login",
                json=json_body.dict(),
                validation_response_code=HTTPStatus.OK,
                return_raw_response=True,
            ),
        )
        return User(**raw_resp.json()), raw_resp.cookies

    def get_current_user(self) -> UserResponse:
        """
        Returns the currently authenticated user

        """

        parsed_resp = cast(
            Dict,
            self.http_client.get("/users/me", validation_response_code=HTTPStatus.OK),
        )
        return UserResponse(**parsed_resp)

    def authenticate(self) -> AuthenticationInfo:
        """
        Returns authentication info for the calling, token-bearing user

        """

        parsed_resp = cast(
            Dict,
            self.http_client.get(
                "/users/me/auth_info", validation_response_code=HTTPStatus.OK
            ),
        )
        return AuthenticationInfo(**parsed_resp)



from bert_score import BERTScorer
from typing import List, Optional
from arthur_bench.scoring import Scorer
from arthur_bench.scoring.utils import suppress_warnings

DEFAULT_MODEL = "microsoft/deberta-v3-base"

# [TODO] need to make these editable by user
DEFAULT_HEDGE = (
    "As an AI language model, I don't have personal opinions, emotions, or beliefs."
)


class HedgingLanguage(Scorer):
    """
    Given an input question and model output, determine if the output contains hedging
    language such as "As an AI language model, I don't have personal opinions, emotions,
    or beliefs". The values returned are a similarity score (BERTScore), with higher
    values corresponding to higher likelihood of hedging language being present in the
    model output.
    """

    @staticmethod
    def name() -> str:
        return "hedging_language"

    @staticmethod
    def requires_reference() -> bool:
        return False

    def __init__(self):
        with suppress_warnings("transformers"):
            self.scorer = BERTScorer(lang="en", model_type=DEFAULT_MODEL)

    def to_dict(self, warn=False):
        return {"model_type": self.scorer.model_type}

    def run_batch(
        self,
        candidate_batch: List[str],
        reference_batch: Optional[List[str]] = None,
        input_text_batch: Optional[List[str]] = None,
        context_batch: Optional[List[str]] = None,
    ) -> List[float]:
        # convert reference hedge to list
        reference_batch = [DEFAULT_HEDGE] * len(candidate_batch)

        # get precision, recall, and F1 score from bert_score package
        p, r, f = self.scorer.score(candidate_batch, reference_batch, verbose=False)

        # return a BERTScore using F1
        list_bertscore = f.tolist()

        return list_bertscore



import uuid
from typing import Optional, List
from arthur_bench.models.models import CreateRunRequest
from arthur_bench.client.bench_client import BenchClient
from arthur_bench.exceptions import ArthurUserError


class TestRun(CreateRunRequest):
    test_suite_id: uuid.UUID
    client: BenchClient  # type: ignore
    id: Optional[uuid.UUID] = None

    class Config:
        arbitrary_types_allowed = True

    @property
    def scores(self) -> List[float]:
        return [case.score for case in self.test_cases]

    @property
    def output(self) -> List[str]:
        return [case.output for case in self.test_cases]

    def save(self) -> uuid.UUID:
        """Save a test run."""
        if self.id is not None:
            raise ArthurUserError("run is already saved")

        resp = self.client.create_new_test_run(
            test_suite_id=str(self.test_suite_id),
            json_body=CreateRunRequest(**self.dict()),
        )
        self.id = resp.id
        return self.id



import os
import duckdb
import getpass
import numpy as np
import glob
import json
from datetime import datetime
from math import ceil
from typing import Optional, Union, List, Dict, Any
from dataclasses import dataclass
import uuid
from pathlib import Path
from arthur_bench.client.bench_client import BenchClient
from arthur_bench.exceptions import NotFoundError, ArthurError, UserValueError
from arthur_bench.models.models import (
    CreateRunRequest,
    CreateRunResponse,
    PaginatedRun,
    PaginatedRuns,
    PaginatedTestSuite,
    PaginatedTestSuites,
    TestSuiteRequest,
    TestSuiteSummary,
    TestSuiteMetadata,
    TestRunMetadata,
    SummaryItem,
    HistogramItem,
    TestCaseRequest,
    TestCaseResponse,
    RunResult,
)

from arthur_bench.utils.loaders import load_suite_from_json, get_file_extension

BENCH_FILE_DIR_KEY = "BENCH_FILE_DIR"
DEFAULT_BENCH_FILE_DIR = Path(os.getcwd()) / "bench_runs"

SUITE_INDEX_FILE = "suite_id_to_name.json"
RUN_INDEX_FILE = "run_id_to_name.json"

TIMESTAMP_FORMAT = "%Y-%m-%dT%H:%M:%S.%f"
DEFAULT_PAGE_SIZE = 5

SORT_QUERY_TO_FUNC = {
    "last_run_time": lambda x: x.last_run_time
    if x.last_run_time is not None
    else x.created_at,
    "name": lambda x: x.name,
    "created_at": lambda x: x.created_at,
    "avg_score": lambda x: x.avg_score,
    "-last_run_time": lambda x: x.last_run_time
    if x.last_run_time is not None
    else x.created_at,
    "-name": lambda x: x.name,
    "-created_at": lambda x: x.created_at,
    "-avg_score": lambda x: x.avg_score,
    "score": lambda x: x.score,
    "id": lambda x: x.id,
}


def _bench_root_dir() -> Path:
    return Path(os.environ.get(BENCH_FILE_DIR_KEY, DEFAULT_BENCH_FILE_DIR))


def _initialize_metadata() -> Dict[str, Any]:
    timestamp = datetime.now().isoformat()
    return {
        "created_at": timestamp,
        "created_by": getpass.getuser(),
        "updated_at": timestamp,
    }


def _load_suite_with_optional_id(
    filepath: Union[str, os.PathLike]
) -> Optional[PaginatedTestSuite]:
    if get_file_extension(filepath) != ".json":
        raise UserValueError("filepath must be json file")
    suite = json.load(open(filepath))
    if "id" in suite:
        return PaginatedTestSuite.parse_obj(suite)
    return None


def _summarize_run(run: PaginatedRun) -> SummaryItem:
    scores = [o.score for o in run.test_cases]
    avg_score = np.mean(scores).item()
    hist, bin_edges = np.histogram(scores, bins=20, range=(0, max(1, np.max(scores))))
    histogram = []
    for i in range(len(hist)):
        hist_item = HistogramItem(
            count=hist[i], low=bin_edges[i], high=bin_edges[i + 1]
        )
        histogram.append(hist_item)
    return SummaryItem(
        id=run.id, name=run.name, avg_score=avg_score, histogram=histogram
    )


@dataclass
class PageInfo:
    sorted_pages: List
    start: int
    end: int
    page: int
    page_size: int
    total_pages: int
    total_count: int


def _paginate(
    objs: List, page: int, page_size: int, sort_key: Optional[str] = None
) -> PageInfo:
    """Paginate sorted files and return iteration indices and page info"""
    if sort_key is not None:
        desc = sort_key[0] == "-"
        sorted_pages = sorted(objs, key=SORT_QUERY_TO_FUNC[sort_key], reverse=desc)
    else:
        sorted_pages = objs
    offset = (page - 1) * page_size
    return PageInfo(
        sorted_pages=sorted_pages,
        start=offset,
        end=min(offset + page_size, len(objs)),
        page=page,
        page_size=page_size,
        total_count=len(objs),
        total_pages=max(1, ceil(len(objs) / page_size)),
    )


class LocalBenchClient(BenchClient):
    """
    Client for managing local file system test suites and runs
    """

    def __init__(self, root_dir: Optional[Union[str, Path]] = None):
        if root_dir is None:
            root_dir = _bench_root_dir()

        # if root dir does not exist, create:
        if not os.path.exists(root_dir):
            os.mkdir(root_dir)

        self.root_dir = Path(root_dir)
        self._write_suite_index()

    def _test_suite_dir(self, test_suite_name: str):
        return Path(self.root_dir) / test_suite_name

    def _create_test_suite_dir(self, test_suite_name: str) -> Path:
        test_suite_dir = self._test_suite_dir(test_suite_name)
        if test_suite_dir.is_dir():
            raise UserValueError(f"test_suite {test_suite_name} already exists")
        os.mkdir(test_suite_dir)
        return test_suite_dir

    def _create_run_dir(self, test_suite_name: str, run_name: str) -> Path:
        run_dir = self._test_suite_dir(test_suite_name) / run_name
        if os.path.exists(run_dir):
            raise UserValueError(f"run {run_name} already exists")
        os.mkdir(run_dir)
        return run_dir

    def _get_suite_name_from_id(self, id: str) -> Optional[str]:
        suite_index = json.load(open(self.root_dir / SUITE_INDEX_FILE))
        if id not in suite_index:
            return None
        return suite_index[id]

    def _get_run_name_from_id(self, test_suite_name: str, id: str) -> Optional[str]:
        run_index = json.load(open(self.root_dir / test_suite_name / RUN_INDEX_FILE))
        if id not in run_index:
            return None
        return run_index[id]

    def _create_test_case_with_id(self, test_case: TestCaseRequest) -> TestCaseResponse:
        return TestCaseResponse(
            id=uuid.uuid4(),
            input=test_case.input,
            reference_output=test_case.reference_output,
        )

    def _update_suite_run_time(self, test_suite_name: str, runtime: datetime):
        suite_file = self.root_dir / test_suite_name / "suite.json"
        suite = PaginatedTestSuite.parse_file(suite_file)
        suite.last_run_time = runtime
        suite.num_runs += 1
        suite_file.write_text(suite.json())

    def _write_suite_index(self):
        suite_index_path = self.root_dir / SUITE_INDEX_FILE
        if suite_index_path.is_file():
            return
        json.dump({}, open(suite_index_path, "w"))
        return None

    def _write_run_index(self, test_suite: str):
        run_index_path = self.root_dir / test_suite / RUN_INDEX_FILE
        if run_index_path.is_file():
            return
        json.dump({}, open(run_index_path, "w"))
        return None

    @staticmethod
    def _update_index(filepath: Path, id: uuid.UUID, name: str):
        suite_index = json.load(open(filepath))
        suite_index[str(id)] = name
        json.dump(suite_index, open(filepath, "w"))

    def _update_suite_index(self, id: uuid.UUID, name: str):
        suite_index_path = self.root_dir / SUITE_INDEX_FILE
        LocalBenchClient._update_index(suite_index_path, id, name)

    def _update_run_index(self, test_suite_name: str, id: uuid.UUID, name: str):
        run_index_path = self.root_dir / test_suite_name / RUN_INDEX_FILE
        LocalBenchClient._update_index(run_index_path, id, name)

    def get_test_suite(
        self, test_suite_id: str, page: int = 1, page_size: int = DEFAULT_PAGE_SIZE
    ) -> PaginatedTestSuite:
        suite_index = json.load(open(self.root_dir / SUITE_INDEX_FILE))
        if test_suite_id not in suite_index:
            raise NotFoundError(f"no test suite with id: {test_suite_id}")
        else:
            suite_file = self.root_dir / suite_index[test_suite_id] / "suite.json"
            suite = PaginatedTestSuite.parse_file(suite_file)
            pagination = _paginate(suite.test_cases, page, page_size)
            return PaginatedTestSuite(
                id=uuid.UUID(test_suite_id),
                name=suite.name,
                scoring_method=suite.scoring_method,
                test_cases=pagination.sorted_pages[pagination.start : pagination.end],
                created_at=suite.created_at,
                updated_at=suite.updated_at,
                last_run_time=suite.last_run_time,
                num_runs=suite.num_runs,
                page=pagination.page,
                page_size=pagination.page_size,
                total_count=pagination.total_count,
                total_pages=pagination.total_pages,
            )

    def get_test_suites(
        self,
        name: Optional[str] = None,
        sort: Optional[str] = None,
        scoring_method: Optional[List[str]] = None,
        page: int = 1,
        page_size: int = DEFAULT_PAGE_SIZE,
    ) -> PaginatedTestSuites:
        # name uniqueness
        if name is not None:
            test_suite_dir = self.root_dir / name
            if test_suite_dir.is_dir():
                suite = _load_suite_with_optional_id(test_suite_dir / "suite.json")
                if suite is None:
                    suite = self.get_test_suite_by_name(name)
                return PaginatedTestSuites(
                    test_suites=[
                        TestSuiteMetadata(
                            id=suite.id,
                            name=suite.name,
                            scoring_method=suite.scoring_method,
                            description=suite.description,
                            created_at=suite.created_at,
                            updated_at=suite.updated_at,
                            last_run_time=suite.last_run_time,
                        )
                    ],
                    page=1,
                    page_size=page_size,
                    total_pages=1,
                    total_count=1,
                )
            else:
                return PaginatedTestSuites(
                    test_suites=[],
                    page=1,
                    page_size=page_size,
                    total_pages=1,
                    total_count=0,
                )

        suites: List[TestSuiteMetadata] = []
        suite_files = glob.glob(f"{self.root_dir}/*/suite.json")
        for f in suite_files:
            suite = _load_suite_with_optional_id(f)
            if suite is None:
                suite = self.get_test_suite_by_name(f.split("/")[-2])
            if scoring_method is None or suite.scoring_method.name in scoring_method:
                suites.append(
                    TestSuiteMetadata(
                        id=suite.id,
                        name=suite.name,
                        scoring_method=suite.scoring_method,
                        description=suite.description,
                        created_at=suite.created_at,
                        updated_at=suite.updated_at,
                        last_run_time=suite.last_run_time,
                    )
                )

        # default sort by last run time
        if sort is None:
            sort = "last_run_time"
        paginate = _paginate(suites, page=page, page_size=page_size, sort_key=sort)
        return PaginatedTestSuites(
            test_suites=paginate.sorted_pages[paginate.start : paginate.end],
            page=paginate.page,
            page_size=paginate.page_size,
            total_pages=paginate.total_pages,
            total_count=paginate.total_count,
        )

    def create_test_suite(self, json_body: TestSuiteRequest) -> PaginatedTestSuite:
        test_suite_dir = self._create_test_suite_dir(json_body.name)
        suite_file = test_suite_dir / "suite.json"

        test_suite_id = uuid.uuid4()
        self._update_suite_index(test_suite_id, json_body.name)
        self._write_run_index(json_body.name)

        resp = PaginatedTestSuite(
            id=test_suite_id,
            name=json_body.name,
            test_cases=[
                self._create_test_case_with_id(test_case)
                for test_case in json_body.test_cases
            ],
            description=json_body.description,
            scoring_method=json_body.scoring_method,
            **_initialize_metadata(),
        )

        suite_file.write_text(resp.json())
        return resp

    def create_new_test_run(
        self, test_suite_id: str, json_body: CreateRunRequest
    ) -> CreateRunResponse:
        test_suite_name = self._get_suite_name_from_id(test_suite_id)
        if test_suite_name is None:
            raise NotFoundError(f"no test suite with id {test_suite_id} ")

        run_id = uuid.uuid4()
        resp = PaginatedRun(
            id=run_id,
            test_suite_id=uuid.UUID(test_suite_id),
            **_initialize_metadata(),
            **json_body.dict(),
        )

        run_dir = self._create_run_dir(test_suite_name, json_body.name)
        run_file = run_dir / "run.json"
        run_file.write_text(resp.json())
        self._update_run_index(test_suite_name, run_id, json_body.name)
        self._update_suite_run_time(
            test_suite_name=test_suite_name, runtime=resp.created_at
        )
        return CreateRunResponse(id=resp.id)

    def get_runs_for_test_suite(
        self,
        test_suite_id: str,
        sort: Optional[str] = None,
        page: int = 1,
        page_size: int = DEFAULT_PAGE_SIZE,
    ) -> PaginatedRuns:
        test_suite_name = self._get_suite_name_from_id(test_suite_id)
        if test_suite_name is None:
            raise NotFoundError(f"no test suite with id: {test_suite_id}")

        runs = []
        run_files = glob.glob(f"{self.root_dir}/{test_suite_name}/*/run.json")
        for f in run_files:
            run_obj = PaginatedRun.parse_file(f)
            avg_score = np.mean([o.score for o in run_obj.test_cases])
            run_resp = TestRunMetadata(**run_obj.dict(), avg_score=float(avg_score))
            runs.append(run_resp)

        if sort is None:
            sort = "created_at"

        pagination = _paginate(runs, page, page_size, sort_key=sort)

        return PaginatedRuns(
            test_runs=pagination.sorted_pages[pagination.start : pagination.end],
            page_size=pagination.page_size,
            page=pagination.page,
            total_pages=pagination.total_pages,
            total_count=pagination.total_count,
        )

    def get_summary_statistics(
        self,
        test_suite_id: str,
        run_id: Optional[str] = None,
        page: int = 1,
        page_size: int = DEFAULT_PAGE_SIZE,
    ) -> TestSuiteSummary:
        test_suite_name = self._get_suite_name_from_id(test_suite_id)
        if test_suite_name is None:
            raise NotFoundError(f"no test suite with id {test_suite_id}")

        runs = []
        run_id_found = False
        run_files = glob.glob(f"{self.root_dir}/{test_suite_name}/*/run.json")
        for f in run_files:
            run_obj = PaginatedRun.parse_file(f)
            runs.append(_summarize_run(run=run_obj))
            if str(run_obj.id) == run_id:
                run_id_found = True

        pagination = _paginate(runs, page, page_size, sort_key="avg_score")
        paginated_summary = TestSuiteSummary(
            summary=runs,
            num_test_cases=len(run_obj.test_cases),
            page_size=pagination.page_size,
            page=pagination.page,
            total_pages=pagination.total_pages,
            total_count=pagination.total_count,
        )

        if run_id is not None and not run_id_found:
            run_name = self._get_run_name_from_id(test_suite_name, run_id)
            if run_name is None:
                raise NotFoundError()
            additional_run = PaginatedRun.parse_file(
                self.root_dir / test_suite_name / run_name / "run.json"
            )
            paginated_summary.summary.append(_summarize_run(additional_run))
        return paginated_summary

    def get_test_run(
        self,
        test_suite_id: str,
        test_run_id: str,
        page: int = 1,
        page_size: int = DEFAULT_PAGE_SIZE,
        sort: Optional[bool] = True,
    ) -> PaginatedRun:
        test_suite_name = self._get_suite_name_from_id(test_suite_id)
        if test_suite_name is None:
            raise NotFoundError(f"test suite with id: {test_suite_id} does not exist")

        run_name = self._get_run_name_from_id(test_suite_name, test_run_id)
        if run_name is None:
            raise NotFoundError(f"test run with id: {test_run_id} does not exist")

        created_at = PaginatedRun.parse_file(
            self.root_dir / test_suite_name / run_name / "run.json"
        ).created_at
        try:
            cases = (
                duckdb.sql(
                    f"SELECT * FROM ("
                    f"SELECT test_cases.id, test_cases.input, "
                    f"test_cases.reference_output FROM ("
                    f"SELECT unnest(test_cases) as test_cases from "
                    f"read_json_auto('{self.root_dir}/{test_suite_name}/suite.json', "
                    f"timestampformat='{TIMESTAMP_FORMAT}'))) "
                    f"POSITIONAL JOIN (SELECT test_cases.output, test_cases.score "
                    f"FROM ("
                    f"SELECT unnest(test_cases) as test_cases from "
                    f"read_json_auto("
                    f"'{self.root_dir}/{test_suite_name}/{run_name}/run.json',"
                    f"timestampformat='{TIMESTAMP_FORMAT}')))"
                )
                .df()
                .to_dict("records")
            )
        except duckdb.IOException:
            cases = []

        run_results = [RunResult.parse_obj(r) for r in cases]
        pagination = _paginate(run_results, page, page_size, sort_key="score")
        return PaginatedRun(
            id=uuid.UUID(test_run_id),
            name=run_name,
            created_at=created_at,
            updated_at=created_at,
            test_case_runs=pagination.sorted_pages[pagination.start : pagination.end],
            test_suite_id=uuid.UUID(test_suite_id),
            page=pagination.page,
            page_size=pagination.page_size,
            total_pages=pagination.total_pages,
            total_count=pagination.total_count,
        )

    def delete_test_suite(self, test_suite_id: str):
        # TODO
        return ArthurError("delete test suite is not supported in local mode yet")

    def delete_test_run(self, test_suite_id: str, test_run_id: str):
        # TODO:
        return ArthurError("delete test run is not supported in local mode yet")

    def get_test_suite_by_name(self, test_suite_name: str) -> PaginatedTestSuite:
        """
        Additional getter to maintain backwards compatibility with non-identified
        local files
        """
        suite_file = self.root_dir / test_suite_name / "suite.json"
        suite = load_suite_from_json(suite_file)

        # override file with index
        id_ = uuid.uuid4()
        resp = PaginatedTestSuite(id=id_, **suite.dict())
        suite_file.write_text(resp.json())
        self._update_suite_index(id_, test_suite_name)
        self._write_run_index(test_suite_name)

        return resp



import logging
import pandas as pd
from typing import List, Optional, Union
from arthur_bench.scoring import Scorer
from arthur_bench.models.models import (
    TestSuiteRequest,
    PaginatedTestSuite,
    TestCaseOutput,
    ScoringMethod,
    ScoringMethodType,
    TestCaseResponse,
)
from arthur_bench.exceptions import (
    UserValueError,
    ArthurInternalError,
)
from arthur_bench.client import BenchClient, _get_bench_client
from arthur_bench.run.testrun import TestRun
from arthur_bench.run.utils import (
    _load_suite_from_args,
    _load_run_data_from_args,
    _initialize_scorer,
)

from arthur_bench.scoring.scorer import SINGLE_ITEM_BATCH_DEFAULT


logger = logging.getLogger(__name__)


class TestSuite:
    """
    Reusable pipeline for running a test suite built from reference_data and evaluated
    using scoring_method

    :param name: name of the test suite
    :param scoring_method: scoring method or scorer instance to use to evaluate the
        results of a test run, as a string/enum or class instance
    :param description: short description of the task tested by this suite
    :param reference_data: dataframe of prompts and reference outputs
    :param reference_data_path: filepath to csv of prompts and reference outputs,
            required if not specifying reference_data
    :param input_column: the column of reference_data containing prompts, defaults to
        'prompt'
    :param reference_column: the column of reference_data containing reference outputs,
        defaults to 'reference'
    :param input_text_list: list of strings of input texts that can be provided instead
        of dataframe columns
    :param reference_output_list: list of strings of reference outputs that can be
        provided instead of dataframe columns
    """

    def __init__(
        self,
        name: str,
        scoring_method: Union[str, Scorer],
        description: Optional[str] = None,
        reference_data: Optional[pd.DataFrame] = None,
        reference_data_path: Optional[str] = None,
        input_column: str = "input",
        reference_column: str = "reference_output",
        input_text_list: Optional[List[str]] = None,
        reference_output_list: Optional[List[str]] = None,
        client: Optional[BenchClient] = None,
    ):
        self._data: PaginatedTestSuite
        if client is None:
            client = _get_bench_client()
        self.client = client
        suite = client.get_suite_if_exists(name=name)
        self.scorer: Scorer

        if suite is None:
            self.scorer = _initialize_scorer(scoring_method_arg=scoring_method)
            cases = _load_suite_from_args(
                reference_data=reference_data,
                reference_data_path=reference_data_path,
                input_column=input_column,
                reference_column=reference_column,
                input_text_list=input_text_list,
                reference_output_list=reference_output_list,
                requires_reference=self.scorer.requires_reference(),
            )
            method_meta = ScoringMethod(
                name=self.scorer.name(),
                type=self.scorer.type(),
                config=self.scorer.to_dict(warn=True),
            )
            new_suite = TestSuiteRequest(
                name=name,
                scoring_method=method_meta,
                description=description,
                test_cases=cases,
            )
            self._data = self.client.create_test_suite(new_suite)

        else:
            logger.info(
                f"Found existing test suite with name {name}. Using existing suite"
            )

            self._data = suite
            if self._data.scoring_method.type == ScoringMethodType.Custom:
                if isinstance(scoring_method, str):
                    raise UserValueError(
                        "cannot reference custom scorer by string. please provide "
                        "instantiated scorer"
                    )

                self.scorer = scoring_method
                if self.scorer.name() != self._data.scoring_method.name:
                    raise UserValueError(
                        f"Test suite was originally created with scorer: "
                        f"{self._data.scoring_method.name}"
                        f"but provided scorer: {scoring_method.name()}"
                    )
                if self.scorer.to_dict() != self._data.scoring_method.config:
                    logger.warning(
                        "scorer configuration has changed from test suite creation."
                    )
            else:
                self.scorer = _initialize_scorer(
                    self._data.scoring_method.name, self._data.scoring_method.config
                )

    @property
    def name(self) -> str:
        return self._data.name

    @property
    def description(self) -> Optional[str]:
        return self._data.description

    @property
    def test_cases(self) -> List[TestCaseResponse]:
        return self._data.test_cases

    @property
    def input_texts(self) -> List[str]:
        return [case.input for case in self._data.test_cases]

    @property
    def reference_outputs(self) -> List[Optional[str]]:
        return [case.reference_output for case in self._data.test_cases]

    @property
    def scoring_method(self) -> str:
        return self.scorer.name()

    def run(
        self,
        run_name: str,
        candidate_data: Optional[pd.DataFrame] = None,
        candidate_data_path: Optional[str] = None,
        candidate_column: str = "candidate_output",
        candidate_output_list: Optional[List[str]] = None,
        context_column: Optional[str] = None,
        context_list: Optional[List[str]] = None,
        save: bool = True,
        batch_size: int = SINGLE_ITEM_BATCH_DEFAULT,
        model_name: Optional[str] = None,
        model_version: Optional[str] = None,
        foundation_model: Optional[str] = None,
        prompt_template: Optional[str] = None,
    ) -> TestRun:
        """
        Score a test run on candidate outputs.

        :param run_name: name for the test run
        :param candidate_data: dataframe of candidate responses to test prompts
        :param candidate_data_path: filepath to csv containing candidate responses to
            test prompts
        :param candidate_column: the column of candidate data containing candidate
            responses, defaults to 'candidate_output'
        :param candidate_output_list: list of strings of candidate outputs that can be
            provided instead of dataframe
        :param context_column: the column of reference_data containing supporting
            context for answering Question & Answering tasks
        :param context_list: list of strings containing supporting context for answering
             question and answering tasks
        :param save: whether to save the run results to file
        :param batch_size: the batch_size to use when computing scores
        :param model_name: model name for model used to generate outputs
        :param model_version: model version of model used to generate outputs
        :param foundation_model: foundation model name used to generate outputs
        :param prompt_template: prompt template name used to generate outputs
        :returns: TestRun object containing scored outputs
        """

        # make sure no existing test run named run_name is already attached to this
        #  suite
        if self.client.check_run_exists(str(self._data.id), run_name):
            raise UserValueError(
                f"A test run with the name {run_name} already exists. "
                "Give this test run a unique name and re-run."
            )

        candidate_output_list, context_list = _load_run_data_from_args(
            candidate_data=candidate_data,
            candidate_data_path=candidate_data_path,
            candidate_column=candidate_column,
            candidate_output_list=candidate_output_list,
            context_column=context_column,
            context_list=context_list,
        )

        if len(candidate_output_list) != len(self.test_cases):
            raise UserValueError(
                f"candidate data has {len(candidate_output_list)} tests but "
                f"expected {len(self.test_cases)} tests"
            )

        inputs = self.input_texts
        ids = [case.id for case in self.test_cases]
        # ref outputs should be None if any items are None (we validate nullness must be
        #  all-or-none)
        ref_outputs: Optional[List[str]] = []
        if ref_outputs is not None:
            for case in self.test_cases:
                if case.reference_output is None:
                    ref_outputs = None
                    break
                else:
                    ref_outputs.append(case.reference_output)
        try:
            all_scores = self.scorer.run(
                candidate_output_list,
                ref_outputs,
                inputs,
                context_list,
                batch_size=batch_size,
            )
        except Exception as e:
            logger.error(f"failed to create run: {e}")
            raise ArthurInternalError(f"failed to create run {run_name}") from e

        test_case_outputs = [
            TestCaseOutput(id=id_, output=output, score=score)
            for id_, output, score in zip(ids, candidate_output_list, all_scores)
        ]

        run = TestRun(
            name=run_name,
            test_case_outputs=test_case_outputs,
            model_name=model_name,
            model_version=model_version,
            foundation_model=foundation_model,
            prompt_template=prompt_template,
            test_suite_id=self._data.id,
            client=self.client,
        )

        if save:
            run.save()

        return run

    def save(self):
        """Save a test suite to local file system."""
        suite_file = self._test_suite_dir / "suite.json"
        suite_file.write_text(self._data.json())
